"""
@brief      test log(time=3s)
"""

import sys
import os
import unittest
import re
import warnings
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pyquickhelper.filehelper import synchronize_folder
from pymyinstall.win_installer.win_patch import win_patch_paths


class TestPatch(unittest.TestCase):

    def test_patch(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if sys.version_info[0] == 2:
            # not setup for python 2.7
            return

        pattern = "([#][!](([A-Za-z][:])?[/\\\\]?[-a-zA-Z0-9_.]+[/\\\\])*pythonw?[.]exe)"
        reg_exe = re.compile(pattern, re.IGNORECASE)
        text = "#!C:\\github\\pymyinstall\\dist\\win_python_setup\\python\\python.exe"
        assert len(reg_exe.findall(text)) > 0

        temp = get_temp_folder(__file__, "temp_patch")
        data = os.path.join(temp, "..", "data", "Scripts")
        dest = os.path.join(temp, "Scripts")
        if not os.path.exists(dest):
            os.mkdir(dest)

        if not os.path.exists(data):
            warnings.warn(
                "pyquickhelper forgot to copy this folder for python 2.7")
            raise FileNotFoundError(data)
        synchronize_folder(data, dest)

        op = win_patch_paths(dest, "", fLOG=fLOG)
        into = "#!python.exe"
        binto = bytes(into, encoding="ascii")
        assert len(op) > 0
        i = 0
        for _, full in op:
            with open(full, "rb") as f:
                content = f.read()
            if binto not in content:
                raise Exception("file:{}\ncontent:\n{}".format(full, content))
            i += 1
        exp = 3
        if i != exp:
            raise Exception("i != " + str(exp) + "\n" + str(op))

        op = win_patch_paths(dest, None, fLOG=fLOG)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=20s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder, ExtTestCase
from pymyinstall.packaged import find_module_install


class TestDownloadPyOpenGL(ExtTestCase):

    def test_install_pyopengl(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if sys.platform.startswith("win"):
            temp = get_temp_folder(__file__, "temp_download_pyopengl")
            m = find_module_install("PyOpenGL_accelerate")
            exe = m.download(
                temp_folder=temp,
                file_save=os.path.join(
                    temp,
                    "out_page.html"), source="2")
            self.assertTrue(os.path.exists(exe))
            self.assertIn("accelerate", m.name)
            down = os.listdir(temp)
            if len(down) != 1:
                raise Exception(down)
            self.assertIn("accelerate", down[0])


if __name__ == "__main__":
    unittest.main()



#!python
"""
script which installs all modules,
works on Linux and Windows if the module is
included the list of modules handled by this module
"""
from __future__ import print_function
import sys
import os
import argparse


def get_parser():
    """
    defines the way to parse the script ``pymy_install``
    """
    typstr = str  # unicode#
    parser = argparse.ArgumentParser(
        description='update modules, consider wheels when the module includes C++ files')
    parser.add_argument(
        '-s',
        '--skip',
        default="",  # "ete,dataspyre,pycuda,cubehelix",
        type=typstr,
        help='list of modules to skip (not to be updated) separated by a comma')
    parser.add_argument(
        '-f',
        '--folder',
        type=typstr,
        default="build/update_modules",
        help='folder where modules will be downloaded')
    parser.add_argument(
        '-d',
        '--deps',
        action='store_true',
        help='install a module or the modules with their dependencies')
    parser.add_argument(
        '--check',
        default="",
        help='check every installed module by importing them, all others options are ignored, ' +
             'the value can be null, a module name or a list comma separated, two integers semi-colon separated 1:end')
    parser.add_argument(
        '--deep-deps',
        action='store_true',
        help='install a module or the modules with their dependencies, check dependencies of dependencies are installed')
    parser.add_argument(
        '--set',
        default="-",
        type=typstr,
        help='set of module to install, see documentation of function get_name_set to get a comprehensive list, ' +
             'this option is ignored if a module is specified on the command line')
    parser.add_argument(
        '--schedule',
        action='store_true',
        help='do not install the modules, returned the list scheduled to be installed')
    parser.add_argument(
        '--force',
        action='store_true',
        help='install or download even the module is already installed')
    parser.add_argument(
        '--download',
        action='store_true',
        help='do not install the modules but download them')
    parser.add_argument(
        '-t',
        '--task',
        default="install",
        type=typstr,
        choices=['install', 'shebang', 'download', 'tool', 'tool_download'],
        help='default task is install but the script can patch shebang in their current location (task=shebang)')
    parser.add_argument(
        'module',
        nargs='*',
        default="all",
        help='update only the list of modules included in this list or all modules if not specified or equal to all')
    parser.add_argument(
        '--source',
        default="",
        type=typstr,
        help='overwrite the source of the wheels')
    return parser


def do_main(temp_folder="build/update_modules",
            skip_module=None,  # ["ete", "dataspyre", "pycuda", "cubehelix"],
            list_module=None, deps=False, schedule_only=False,
            deep_deps=False, checkings=None,
            task="install", source=None, download_only=False,
            force=False):
    """
    calls function @see fn install_all but is meant to be added to scripts folder

    @param      temp_folder     folder where modules will be downloaded
    @param      skip_module     skip the module on this list
    @param      list_module     list of modules to update or None for all
    @param      deps            install a module with its dependencies
    @param      schedule_only   if True, the function returns the list of modules scheduled to be installed
    @param      deep_deps       check dependencies for dependencies
    @param      checkings       if True, run checkings, do not install anything, example of values
    @param      task            *install* or *shebang* or *download*
                                ``""``, ``matplotlib``, ``100,end``,
                                option *download* is equivalent to *checkings*
    @param      source          overwrite the source of the wheel
    @param      download_only   only download the modules, no installation
    @param      force           force the download or the installation

    If *deps* is True, *list_module* cannot be empty.

    .. versionchanged:: 1.1
        Parameters *source*, *force* were added.
    """
    try:
        from pymyinstall import is_travis_or_appveyor
    except ImportError:
        def is_travis_or_appveyor():
            if "travis" in sys.executable:
                return "travis"
            if os.environ.get("USERNAME", os.environ.get("USER", None)) == "appveyor":
                return "appveyor"
            return None
    if is_travis_or_appveyor() and source is None:
        source = "2"

    if task in ("install", "download"):
        checkings = checkings or task == "download"
        if checkings:
            try:
                from pymyinstall.win_installer import import_every_module
            except ImportError:
                pfolder = os.path.normpath(os.path.join(
                    os.path.abspath(os.path.dirname(__file__)), "..", ".."))
                sys.path.append(pfolder)
                if "pymyinstall" in sys.modules:
                    del sys.modules["pymyinstall"]
                if "pymyinstall.win_installer" in sys.modules:
                    del sys.modules["pymyinstall.win_installer"]
                from pymyinstall.win_installer import import_every_module

            def to_int(s):
                if "end" in s:
                    return -1
                try:
                    return int(s)
                except ValueError:
                    return -1

            if checkings in ("", "0,end", "0,-1"):
                module_list = None
                start = 0
                end = -1
            elif ":" in checkings:
                module_list = None
                spl = checkings.split(":")
                if len(spl) == 2:
                    start = to_int(spl[0])
                    end = to_int(spl[1])
                else:
                    raise ValueError("unable to interpret: " + checkings)
            else:
                module_list = [_.strip() for _ in checkings.split(",")]
                start = 0
                end = -1

            print("CHECKINGS {}:{} -- {}".format(start, end,
                                                 "all" if module_list is None else ",".join(module_list)))
            import_every_module(None, module_list, start=start, end=end)
        else:
            if not os.path.exists(temp_folder):
                os.makedirs(temp_folder)
            try:
                from pymyinstall.packaged import install_all
            except ImportError:
                folder = os.path.normpath(os.path.join(
                    os.path.abspath(os.path.dirname(__file__)), "..", ".."))
                sys.path.append(folder)
                if "pymyinstall" in sys.modules:
                    del sys.modules["pymyinstall"]
                if "pymyinstall.packaged" in sys.modules:
                    del sys.modules["pymyinstall.packaged"]
                from pymyinstall.packaged import install_all
            res = install_all(temp_folder=temp_folder, verbose=True,
                              skip_module=skip_module, list_module=list_module, deps=deps,
                              schedule_only=schedule_only,
                              deep_deps=deep_deps, source=source,
                              download_only=download_only, force=force)
            if schedule_only:
                print("SCHEDULED")
                for r in res:
                    print(r)
    elif task == "shebang":
        try:
            from pymyinstall.win_installer import win_patch_paths
            from pymyinstall.installhelper import get_pip_program
        except ImportError:
            pfolder = os.path.normpath(os.path.join(
                os.path.abspath(os.path.dirname(__file__)), "..", ".."))
            sys.path.append(pfolder)
            if "pymyinstall" in sys.modules:
                del sys.modules["pymyinstall"]
            if "pymyinstall.packaged" in sys.modules:
                del sys.modules["pymyinstall.packaged"]
            if "pymyinstall.win_installer" in sys.modules:
                del sys.modules["pymyinstall.win_installer"]
            from pymyinstall.win_installer import win_patch_paths
            from pymyinstall.installhelper import get_pip_program
        pip = get_pip_program()
        folder = os.path.dirname(os.path.abspath(pip))
        win_patch_paths(temp_folder, None, fLOG=print)
    elif task in ("tool", "tool_download"):
        try:
            from pymyinstall.installcustom import install_graphviz
        except ImportError:
            pfolder = os.path.normpath(os.path.join(
                os.path.abspath(os.path.dirname(__file__)), "..", ".."))
            sys.path.append(pfolder)
            if "pymyinstall" in sys.modules:
                del sys.modules["pymyinstall"]
            if "pymyinstall.installcustom" in sys.modules:
                del sys.modules["pymyinstall.installcustom"]
            from pymyinstall.installcustom import install_graphviz
        if list_module is None:
            raise ValueError("A tool must be precised, list cannot be empty.")
        low = [_.lower() for _ in list_module]
        for tool in low:
            if tool == "graphviz":
                install_graphviz(temp_folder, install=task ==
                                 "tool", fLOG=print, source=source)
            else:
                raise NameError("unable to install '{0}'".format(tool))
    else:
        raise ValueError("unable to interpret task: " + task)


def main():
    """
    calls function @see fn install_all but is meant to be added to scripts folder,
    parse command line arguments
    """
    parser = get_parser()
    try:
        res = parser.parse_args()
    except SystemExit:
        print(parser.format_usage())
        res = None

    if res is not None:
        skip_module = None if res.skip in [
            "", "-", None, []] else res.skip.split(",")
        list_module = None if res.module in [
            "all", "", "-", None, []] else res.module
        if list_module is None and res.set is not None and len(res.set) > 0 and res.set != "-":
            list_module = res.set
        do_main(temp_folder=res.folder, skip_module=skip_module,
                list_module=list_module, deps=res.deps, schedule_only=res.schedule,
                deep_deps=res.deep_deps, checkings=res.check, task=res.task,
                source=res.source if res.source else None,
                download_only=res.download, force=res.force)


if __name__ == "__main__":
    main()



"""
@brief      test log(time=45s)
"""
import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installcustom import install_chromedriver


class TestChromeDriver(unittest.TestCase):

    def test_install_ChromeDriver(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_chromedriver")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            f = os.path.join(temp, _)
            if os.path.isfile(f):
                os.remove(f)

        r = install_chromedriver(temp, fLOG=fLOG, install=True)
        for _ in r:
            fLOG(_)
        if sys.platform.startswith("win"):
            found = os.path.join(temp, "chromedriver.exe")
            if not os.path.exists(found):
                raise FileNotFoundError(found)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
*pymyinstall* installs package on Windows and Linux.
It automatically chooses a location for a
:ref:`long list of package <l-ensae_fullset-table>`.
To install a list of modules for a machine learner:

::

    from pymyinstall import complete_installation, install_scite, install_pandoc, open_tool_on_browser
    for _ in complete_installation():
        _.install(temp_folder="install")

You can also install some useful tools:

::

    from pymyinstall import install_scite, install_pandoc

    install_scite("install")
    install_pandoc("install")

To download a module:

::

    from pymyinstall import download_module
    download_module("module_name")

To install a module:

::

    from pymyinstall import install_module
    install_module("module_name")
"""
from .installhelper.install_cmd_helper import run_cmd, unzip_files
from .installhelper.module_install import ModuleInstall
from .installcustom.install_custom import download_from_sourceforge, download_file, download_page
from .installhelper.install_manual import get_install_list
from .installhelper import get_module_version, get_pypi_version
from .installcustom.install_custom_revealjs import download_revealjs
from .installhelper.requirements import build_requirements
from .win_installer.win_setup_main import win_python_setup
from .packaged import install_module, update_module, download_module


__version__ = "1.4.1956"
__author__ = "Xavier Dupré"
__github__ = "https://github.com/sdpython/pymyinstall"
__url__ = "http://www.xavierdupre.fr/app/pymyinstall/helpsphinx/index.html"
__license__ = "MIT License"


def _setup_hook():
    """
    does nothing
    """
    # we clean the cache
    ModuleInstall.clear_cache()


def check(log=False):
    """
    Checks the library is working.
    It raises an exception.
    If you want to disable the logs:

    @param      log     if True, display information, otherwise
    @return             0 or exception
    """
    return True


def is_travis_or_appveyor():
    """
    tells if is a travis environment or appveyor

    :return:        travis, appveyor or None

    .. versionadded:: 1.1
    """
    import sys
    if "travis" in sys.executable:
        return "travis"
    import os
    if os.environ.get("USERNAME", os.environ.get("USER", "")) == "appveyor":
        return "appveyor"
    return None



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install import ModuleInstall
from pymyinstall.installhelper.install_cmd_helper import run_cmd


class TestInstall (unittest.TestCase):

    def test_run_cmd(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        out, err = run_cmd("find", wait=True, fLOG=fLOG)
        self.assertTrue(out is not None)
        self.assertTrue(err is not None)

    def test_install(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        m = ModuleInstall("pip", "pip")
        self.assertTrue(m.install())
        m = ModuleInstall("pip", "exe")
        self.assertTrue(m.install())


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=5s)
"""

import os
import unittest
from pyquickhelper.loghelper import fLOG, CustomLog
from pyquickhelper.pycode import is_travis_or_appveyor
from pymyinstall.packaged import small_set
from pymyinstall.installhelper import get_wheel_version, compare_version
from pymyinstall.installhelper.module_install_exceptions import MissingWheelException


class TestCompareVersion(unittest.TestCase):

    def test_compare_versions(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if is_travis_or_appveyor():
            return

        mods = [mod for mod in small_set() if mod.kind in ("wheel",)]
        fLOG(len(mods))
        self.assertTrue(len(mods) > 0)
        clog = CustomLog(os.path.dirname(__file__),
                         "windows_modules_versions.txt")

        diff = []
        for i, mod in enumerate(mods[0:]):
            if mod.name in ["PySide", "cgal_bindings", "ad3"]:
                # skipping modules not available
                continue

            if i % 10 == 0:
                fLOG("[test_compare_versions]", i, '-', mod)
            try:
                url1 = mod.get_exewheel_url_link(wheel=True)
            except MissingWheelException:
                url1 = None
            try:
                url2 = mod.get_exewheel_url_link2(wheel=True, source="2")
            except MissingWheelException:
                url2 = None
            if url1 is None and url2 is None:
                fLOG("[test_compare_versions] missing package on both sides", mod)
                clog("[test_compare_versions] missing package on both sides", mod)
                continue

            n1 = url1[-1] if url1 is not None else None
            n2 = url2[-1] if url2 is not None else None
            v1 = get_wheel_version(n1) if n1 is not None else None
            v2 = get_wheel_version(n2) if n2 is not None else None
            # might be just a date
            if v1 is not None and '.' not in v1 and len(v1) != 8:
                raise ValueError(
                    "Wrong version number (1) '{0}' for '{1}' - version '{2}'".format(v1, mod.name, n1))
            # might be just a date
            if v2 is not None and '.' not in v2 and len(v2) != 8:
                raise ValueError(
                    "Wrong version number (2) '{0}' for '{1}' - version '{2}'".format(v2, mod.name, n2))
            cmp = compare_version(
                v1, v2) if v1 is not None and v2 is not None else -2
            if cmp != 0:
                fLOG("[test_compare_versions] ---", i, '-', v1, v2, mod.name)
                clog("[test_compare_versions] ---", i, '-', v1, v2, mod.name)
                if v1 is not None and v2 is not None and not ("rc1" in v1 or "rc1" in v2):
                    diff.append(mod)

        self.assertTrue(len(diff) <= 10)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=20s)
"""
import os
import unittest


class TestDownloadSetup (unittest.TestCase):

    def test_setup(self):
        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_download_setup")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            if os.path.isfile(os.path.join(temp, _)):
                os.remove(os.path.join(temp, _))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
@file
@brief Functions get_page_wheel
"""

import sys
from ssl import SSLEOFError
from .install_memoize import install_memoize
from .internet_settings import default_user_agent

if sys.version_info[0] == 2:
    import urllib2 as urllib_request
    from codecs import open
    from HTMLParser import HTMLParser
else:
    import urllib.request as urllib_request
    from html.parser import HTMLParser
    from urllib.error import URLError


class InternalJsException(Exception):
    """
    Raises when a javascript url cannot be decrypted.
    """
    pass


@install_memoize
def get_page_wheel(page, sele=True):
    """
    get the page

    @param      page        location
    @param      sele        use selenium or not or False to try if the other way did not work
    @return                 page content
    """
    req = urllib_request.Request(
        page,
        headers={
            'User-agent': default_user_agent})
    ull = False
    try:
        u = urllib_request.urlopen(req)
        ull = True
    except (SSLEOFError, URLError) as ee:
        # This usually happens on Windows.
        # ssl.SSLEOFError: EOF occurred in violation of protocol (_ssl.c:749)
        if sele:
            from ..installcustom.install_custom_chromedriver import install_chromedriver
            import selenium.webdriver
            install_chromedriver(fLOG=None)
            browser = selenium.webdriver.Chrome()
            browser.get(page)
            text = browser.page_source
            browser.close()
            if len(text) < 1000:
                raise ValueError(
                    "Unable to retrieve information from '{0}' with selenium len={1}".format(page, len(text)))
        else:
            raise ee
    except Exception as e:
        raise Exception(
            "unable to get '{0}' '{1}'".format(page, type(e))) from e

    if ull:
        text = u.read()
        u.close()
        text = text.decode("utf8")

    return _clean_page_wheel(text)


def _clean_page_wheel(text):
    """
    remove unexpected characters

    @param      text        string
    @return                 string
    """
    text = text.replace("&quot;", "'")
    text = text.replace("&#8209;", "-")
    text = text.replace("&#46;", ".")
    text = text.replace(" &middot; ", "-")
    text = text.replace("&ndash;", "-")
    return text


def save_page_wheel(filename, content):
    """
    cache a HTML page

    @param      filename        filename
    @param      content         content
    @return                     filename
    """
    with open(filename, "w", encoding="utf8") as f:
        f.write(content)


def read_page_wheel(filename):
    """
    read a cached HTML page

    @param      filename        filename
    @return                     filename
    """
    with open(filename, "r", encoding="utf8") as f:
        text = f.read()
    return _clean_page_wheel(text)


def _cg_dl1(ml, mi):
    ot = ""
    for j in range(0, len(mi)):
        ot += chr(ml[ord(mi[j]) - 48])
    return ot


def _cg_dl(ml, mi, fLOG=None):
    """
    compressed::

        if (top.location!=location) top.location.href=location.href;
        function dc(ml,mi){var ot="";for(var j=0;j<mi.length;j++)ot+=String.fromCharCode(ml[mi.charCodeAt(j)-48]);
        document.write(ot);}function dl1(ml,mi){var ot="";for(var j=0;j<mi.length;j++)ot+=String.fromCharCode(ml[mi.charCodeAt(j)-48]);
        location.href=ot;}function dl(ml,mi){mi=mi.replace('&lt;','<');mi=mi.replace('&#62;','>');mi=mi.replace('&#38;','&');
        setTimeout(function(){dl1(ml,mi)},1500);}

    source::

        <script type="text/javascript">
        // <![CDATA[
        if (top.location!=location)
            top.location.href=location.href;
        function dc(ml,mi)
        {
            var ot="";
            for(var j=0;j<mi.length;j++)
                ot+=String.fromCharCode(ml[mi.charCodeAt(j)-48]);
            document.write(ot);
        }
        function dl1(ml,mi)
        {
            var ot="";
            for(var j=0;j<mi.length;j++)
                ot+=String.fromCharCode(ml[mi.charCodeAt(j)-48]);
                location.href=ot;
        }
        function dl(ml,mi)
        {
            mi=mi.replace('&lt;','<');
            mi=mi.replace('&#62;','>');
            mi=mi.replace('&#38;','&');
            setTimeout(function(){dl1(ml,mi)},1500);
        }
        // ]]>
        </script>
    """
    if fLOG:
        fLOG("[pymy] decode", ml)
        fLOG("[pymy] decode", mi)
    mi = mi.replace('&lt;', '<')
    mi = mi.replace('&#62;', '>')
    mi = mi.replace('&gt;', '>')
    mi = mi.replace('&#38;', '&')
    return _cg_dl1(ml, mi)


class HTMLParser4Links(HTMLParser):
    """
    extreact all links ni HTML page
    """

    def __init__(self):
        """
        constructor
        """
        if sys.version_info[0] == 2:
            HTMLParser.__init__(self)
        else:
            HTMLParser.__init__(self, convert_charrefs=True)
        self.links = []
        self.current = None

    def handle_starttag(self, tag, attrs):
        """
        enters a tag
        """
        if tag == "a":
            self.current = ""
            self.attrs = attrs

    def handle_endtag(self, tag):
        """
        ends of a tag
        """
        def clean_dashes(st):
            b = st.encode('utf-8')
            b = b.replace(b'\xe2\x80\x91', b'-')
            b = b.replace(b'\xc2\xa0', b' ')
            return b.decode('utf-8')
        if tag == "a":
            if self.current is not None and len(self.current) > 0:
                app = (clean_dashes(self.current),
                       [(clean_dashes(name), clean_dashes(link)) for name, link in self.attrs])
                self.links.append(app)
            self.current = None

    def handle_data(self, data):
        """
        stores data if a link
        """
        if self.current is not None:
            self.current += data


def extract_all_links(text):
    """
    parses HTML to extract all links

    @param      text        HTML page
    @return                 list of links
    """
    parser = HTMLParser4Links()
    parser.feed(text)
    return parser.links


def enumerate_links_module(name, alls, version, plat):
    """
    Selects the links for a specific module.

    @param      name        module name
    @param      alls        all links from @see fn extract_all_links
    @param      version     python version
    @param      plat        platform
    """
    version = "%d%d" % version[:2]
    lname = name.lower()
    lname_ = lname.replace("-", "_") + "-"
    lname += "-"
    for a in alls:
        n = a[0]
        ln = n.lower()
        if (ln.startswith(lname) or ln.startswith(lname_)) and plat in ln:
            vers = ("cp" + version, "py" + version)
            good = False
            for v in vers:
                if v in ln:
                    good = True
            if not good:
                continue
        else:
            continue

        js = None
        for at, val in a[1]:
            if at == "onclick":
                js = val.lstrip()

        if js:
            js0 = js
            suf = '"javascript:dl("'
            bs = ["javascript:", "javascript :", "javascript  :"]
            res = None
            for b in bs:
                if js.startswith(b):
                    js = js[len(b):]
                    if js.endswith(suf):
                        js = js[:-len(suf) - 2]
                    if "javascript:" in js:
                        # Addition: 207-08-24
                        js = js[:js.index('javascript:')]
                    dl = _cg_dl
                    js = js.strip('" \t ;\'')
                    if dl is not None:
                        try:
                            res = eval(js)
                        except SyntaxError as e:
                            raise SyntaxError(
                                "Unable to evaluate '{0}'\njs0='{1}'.".format(js, js0)) from e
                        break
            if res is None:
                raise InternalJsException(
                    "Unable to decode js '{0}'".format(js))
            yield n, js, res



"""
@brief      test log(time=45s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installcustom import install_tdm_gcc


class TestTDMGCC (unittest.TestCase):

    def test_install_tdm_gcc(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if sys.version_info[0] == 2:
            return

        temp = get_temp_folder(__file__, "temp_tdc_gcc")

        if sys.platform.startswith("win"):
            r = install_tdm_gcc(temp, fLOG=fLOG)
            exe = os.path.abspath(r)
            self.assertTrue(os.path.exists(exe))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=12s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG, CustomLog
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installcustom import install_python, folder_older_than


class TestDownloadPythonBase(unittest.TestCase):

    def test_install_python_base(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        vers = "%d%d" % sys.version_info[:2]
        temp = get_temp_folder(__file__, "temp_py%s_base" %
                               vers, persistent=True, clean=folder_older_than)
        down = get_temp_folder(
            __file__, "temp_py%s_base_download" % vers, clean=True, persistent=True)
        self.assertEqual(len(os.listdir(down)), 0)

        clog = CustomLog(temp) if __name__ != "__main__" else fLOG
        install_python(install=True, temp_folder=temp, fLOG=clog,
                       download_folder=temp + "_download")
        if sys.platform.startswith("win"):
            pyt = os.path.join(temp, "python.exe")
            pip = os.path.join(temp, "Scripts", "pip.exe")
            if not os.path.exists(pyt):
                raise FileNotFoundError(pyt)
            if not os.path.exists(pip):
                raise FileNotFoundError(pip)
        else:
            # already checked
            pass


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=63s)
"""

import sys
import os
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import is_travis_or_appveyor
from pymyinstall.installhelper.install_cmd_helper import run_cmd


class TestPyMyUpdateCli(unittest.TestCase):

    def test_update_set_schedule(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if is_travis_or_appveyor() == "travis":
            # run_cmd no end on travis.
            return
        this = os.path.abspath(os.path.dirname(__file__))
        script = os.path.normpath(os.path.join(
            this, "..", "..", "src", "pymyinstall", "cli", "pymy_update.py"))
        cmd = "{0} -u {1} {2}".format(
            sys.executable, script, "--schedule numpy matplotlib")
        out, err = run_cmd(cmd, wait=True)
        if len(out) == 0:
            if is_travis_or_appveyor() == "appveyor":
                warnings.warn(
                    "CLI ISSUE cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
            else:
                raise Exception(
                    "cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=3s)
"""
import sys
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall import is_travis_or_appveyor
from pymyinstall.win_installer.win_setup_main_checkings import distribution_checkings


class TestCheckings(unittest.TestCase):

    def test_checkings(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        if "Anaconda custom" in sys.version:
            try:
                distribution_checkings(None, None, fLOG=fLOG, skip_import=True)
            except Exception as e:
                # We skip errors involving spyder and rodeo as unit test are run from a virtual
                # environment and they might be present.
                lines = str(e).split()
                for line in lines:
                    if "pymyinstall" in line:
                        # if "rodeo" not in line and "spyder" not in line:
                        # rodeo is not checked anymore, the installation
                        # changed
                        if "spyder" not in line:
                            raise Exception(
                                "spyder not found in line\n{0}".format(line)) from e
        elif not is_travis_or_appveyor():
            try:
                distribution_checkings(None, None, fLOG=fLOG, skip_import=True)
            except Exception as e:
                if '_venv' not in str(e) or ('Scripts' not in str(e) and '.exe' not in str(e)):
                    raise Exception("version: " + sys.version) from e


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=100s)
"""
import sys
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.win_installer import import_every_module


class TestSuccessfulCustom(unittest.TestCase):

    def test_long_import_every_module(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        subset = ["xlrd"]
        subset.extend("docutils pandas numpy".split())
        res = import_every_module(
            sys.executable, subset, fLOG=fLOG)
        nb = 0
        for r in res:
            if not r[0]:
                fLOG("---------------------------------")
                fLOG("FAILED", r[1], "\nOUT\n", r[2], "\nERR\n", r[3])
            else:
                nb += 1
        assert nb > 0


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=45s)
"""
import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installcustom import install_scala_sbt, ManualDownloadException


class TestScalaSbt(unittest.TestCase):

    @unittest.skipIf(not sys.platform.startswith("win"), reason="not implemented on linux")
    def test_install_scala_sbt(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_scala_sbt")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            if os.path.isfile(os.path.join(temp, _)):
                os.remove(os.path.join(temp, _))

        try:
            r = install_scala_sbt(temp_folder=temp, fLOG=fLOG, install=False)
            fLOG(r)
        except ManualDownloadException:
            pass


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install_version import get_wheel_version


class TestInstallHelper(unittest.TestCase):

    def test_version_name(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        v = get_wheel_version("rpy2-2.6.0-cp34-none-win_amd64.whl")
        if v != "2.6.0":
            raise Exception(v)

        v = get_wheel_version("tornado-4.2-cp34-none-win_amd64.whl")
        if v != "4.2":
            raise Exception(v)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=7s)
"""

import sys
import os
import unittest
import warnings
import subprocess
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import is_travis_or_appveyor
from pymyinstall.installhelper.install_cmd_helper import run_cmd


class TestPyMyDepsCli(unittest.TestCase):

    def test_install_deps(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        this = os.path.abspath(os.path.dirname(__file__))
        script = os.path.normpath(os.path.join(
            this, "..", "..", "src", "pymyinstall", "cli", "pymy_deps.py"))
        cmd = "{0} -u {1} {2}".format(
            sys.executable, script, "pandas")
        try:
            out, err = run_cmd(cmd, wait=True, fLOG=fLOG,
                               communicate=True, timeout=120)
        except subprocess.CalledProcessError as e:
            mes = "CMD\n{0}\nOUT\n{1}\nERR\n{2}".format(
                e.cmd, e.output, e.stderr.read() if e.stderr else "")
            raise Exception(mes) from e
        out = out.strip()
        if len(out) == 0:
            if is_travis_or_appveyor() == "appveyor":
                warnings.warn(
                    "CLI ISSUE cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
            else:
                raise Exception(
                    "cmd:\n{0}\nOUT:\n{1}\nERR\n{2}".format(cmd, out, err))
        else:
            fLOG(out)
            if sys.version_info[0] == 2:
                # failing on Python 2.7
                return
            if "pandas" not in out:
                if len(err) == 0 and sys.version_info[:2] == (3, 6):
                    # I don't know why this test is failing.
                    return
                raise Exception(
                    "(1)CMD:\n{0}\nOUT\n{1}\nERR\n{2}".format(cmd, out, err))
            if "['pandas']" not in out:
                if len(err) == 0 and sys.version_info[:2] == (3, 6):
                    # I don't know why this test is failing.
                    return
                raise Exception(
                    "(2)CMD:\n{0}\nOUT\n{1}\nERR\n{2}".format(cmd, out, err))


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Various function to install various python module from various location.
"""
from __future__ import print_function
import sys
import re
import os
import time
import importlib
import datetime
import warnings
from urllib.parse import urlsplit
import urllib.request as urllib_request
import urllib.error as urllib_error
import importlib.util
import xmlrpc.client as xmlrpc_client
from pip import __version__ as pip_version
from .install_cmd_helper import python_version, run_cmd, unzip_files, get_pip_program
from .install_cmd_helper import get_python_program, get_file_modification_date, get_conda_program, is_conda_distribution
from .module_install_exceptions import MissingPackageOnPyPiException, MissingInstalledPackageException, InstallError
from .module_install_exceptions import DownloadError, MissingVersionWheelException, WrongWheelException, MissingWheelException
from .module_install_version import get_pypi_version, get_module_version, annoying_modules, get_module_metadata
from .module_install_version import numeric_version, compare_version, choose_most_recent, get_wheel_version
from .module_install_page_wheel import get_page_wheel, read_page_wheel, save_page_wheel, enumerate_links_module, extract_all_links
from .missing_license import missing_module_licenses
from .internet_settings import default_user_agent
from .install_cmd_regex import regex_wheel_versions


def _filter_pip_out(out):
    lines = out.split('\n')
    res = []
    for line in lines:
        if "WARNING:" in line:
            continue
        if "Consider adding this directory to PATH" in line:
            continue
        res.append(line)
    return "\n".join(res).strip(' \n\r\t')


class ModuleInstall:

    """
    defines the necessary information for a module

    .. exref::
        :title: Installation from GitHub

        ::

            ModuleInstall("pyquickhelper", "github",
                          "sdpython").install(temp_folder="temp")
    """

    allowedKind = ["pip", "github", "exe", "exe2", "wheel", "wheel2"]
    exeLocation = "https://www.lfd.uci.edu/~gohlke/pythonlibs/"
    exeLocationXd_Default = "http://www.xavierdupre.fr/enseignement/setup/"
    gitexe = r"C:\Program Files (x86)\Git"
    github_pattern_zip = "https://github.com/{1}/{0}/archive/{2}.zip"
    github_pattern_git = "https://github.com/{1}/{0}.git{2}"

    @staticmethod
    def is_annoying(module_name):
        """
        some modules are not available on pipy
        """
        return module_name in annoying_modules

    def __init__(self, name, kind="pip", gitrepo=None, mname=None, fLOG=print,
                 version=None, script=None, index_url=None, deps=None,
                 purpose=None, usage=None, web=None, source=None, custom=None,
                 branch="master", pip_options=None, overwrite=None, post=None,
                 skip_import=False, pipgit=False):
        """
        @param      name            name
        @param      kind            kind of installation (*pip*, *github*, *wheel*)
        @param      gitrepo         github repository (example: sdpython)
        @param      mname           sometimes, the module name is different from its official name
        @param      version         to install a specific version (None for the latest)
        @param      fLOG            logging function
        @param      script          some extensions are not a module but an application (such as ``spyder``),
                                    the class will check this script is available
        @param      deps            overwrite deps parameters when installing the module
        @param      index_url       to get the package from a custom pypi server
        @param      purpose         purpose of the module
        @param      usage           main usage for the module
        @param      web             website for the module, if None, default to pipy
        @param      source          to overwrite parameter *source* of methods
                                    @see me download, @see me install or @see me update.
        @param      custom          custom instructions to install, usually
                                    ``['build', 'install']`` to run
                                    ``setup.py build`` and ``setup.py install``
        @param      branch          only necessary for install process with github
        @param      pip_options     additional options for pip (list)
        @param      overwrite       overwrite the location of the wheel
        @param      post            instructions post installation (look for this parameter
                                    in the code to see what is supported)
        @param      pipgit          install the module with ``pip + git`` instead of
                                    getting the full archive
        @param      skip_import     added to indicate the module cannot be imported

        .. versionchanged:: 1.1
            Parameters *source*, *custom*, *branch*, *pip_options*, *overwrite*, *post* were added.
            Parameter *skip_import* was introduced to skip the checking of the installation.
            Parameter *pipgit* was added.
        """
        if kind != "pip" and version is not None:
            raise NotImplementedError(
                "version can be only specified if kind=='pip'")

        self.name = name
        self.kind = kind
        self.gitrepo = gitrepo
        self.version = version
        self.mname = mname
        self.script = script
        self.index_url = index_url
        self.deps = deps
        self.purpose = purpose
        self.usage = usage
        self.existing_version = None
        self.source = source
        self.custom = custom
        self.branch = branch
        self.pip_options = pip_options
        self.overwrite = overwrite
        self.post_installation = post
        self.pipgit = pipgit

        if self.mname == self.name:
            raise ValueError(
                "Do not specify mname if it is equal to name '{0}'.".format(self.name))
        if self.pipgit and self.gitrepo is None:
            raise ValueError("If pipgit=True, gitrepo must be specified.")
        if self.pipgit and kind != 'github':
            raise ValueError(
                "If pipgit=True, kind must be 'github' not '{0}'.".format(kind))
        if self.version is not None and self.gitrepo is not None:
            raise ValueError("version must be None if gitrepo is not.")

        if self.mname is not None and self.mname.startswith("-"):
            self.mname = self.mname[1:]
            self.skip_import = True
        else:
            self.skip_import = skip_import
        self.web = web if web is not None else (
            "https://pypi.python.org/pypi/" + self.name)

        if self.kind not in ModuleInstall.allowedKind:
            raise Exception(
                "unable to interpret kind {0}, it should be in {1}".format(
                    kind, ",".join(
                        ModuleInstall.allowedKind)))
        if self.kind == "github" and self.gitrepo is None:
            raise Exception("gitrepo cannot be empty")

        self.fLOG = fLOG

    def copy(self, version=None):
        """
        copy the module, if version is not None, change the version number

        @param      version     version number or None for unchanged
        @return                 @see cl ModuleInstall

        .. versionadded:: 1.0
        """
        mod = ModuleInstall(**self.as_dict())
        if version is not None:
            mod.version = version
        return mod

    def as_dict(self, rst_link=False):
        """
        returns the members in a dictionary

        @param      rst_link    if True, add rst_link, license, classifier
        @return                 dictionary
        """
        r = dict(name=self.name, kind=self.kind, gitrepo=self.gitrepo, version=self.version,
                 mname=self.mname, script=self.script, deps=self.deps, index_url=self.index_url,
                 purpose=self.purpose, usage=self.usage, web=self.web,
                 post=None if self.post_installation is None else self.post_installation.copy(),
                 skip_import=self.skip_import, pipgit=self.pipgit)
        if rst_link:
            r["rst_link"] = "`{0} <{1}>`_".format(self.name, self.web)
            r["license"] = self.get_installed_license()
            r["installed"] = self.get_installed_version()
            r["classifier"] = self.get_installed_classifier()
        return r

    def __cmp__(self, o):
        """
        to sort modules

        @param      o       other module
        @return             -1, 0, 1
        """
        def compare(v1, v2):
            if v1 is None:
                if v2 is None:
                    return 0
                else:
                    return 1
            else:
                if v2 is None:
                    return -1
                else:
                    if v1 < v2:
                        return -1
                    elif v1 > v2:
                        return 1
                    else:
                        return 0

        r = compare(self.usage, o.usage)
        if r != 0:
            return r
        r = compare(self.name.lower(), o.name.lower())
        return r

    def __lt__(self, o):
        """
        overload operator ``<``

        @param      o       other module
        @return             boolean
        """
        return self.__cmp__(o) < 0

    @staticmethod
    def clear_cache():
        """
        clear the local cache to get wheel link
        """
        if os.path.exists(ModuleInstall._page_cache_html):
            os.remove(ModuleInstall._page_cache_html)
        if os.path.exists(ModuleInstall._page_cache_html2):
            os.remove(ModuleInstall._page_cache_html2)

    @property
    def Purpose(self):
        """
        returns the comment
        """
        return "-" if self.purpose is None else self.purpose

    @property
    def Script(self):
        """
        returns the script to run if the extension is an application and not a module
        """
        exe = os.path.split(sys.executable)[0]
        if sys.platform.startswith("win"):
            sc = os.path.join(exe, "Scripts", self.script)
        else:
            sc = os.path.join(exe, "Scripts", os.path.splitext(self.script)[0])
        return sc

    def __str__(self):
        """
        usual
        """
        if self.script is None:
            return "{0}:{1}:import {2}:v{3}".format(
                self.name, self.kind, self.ImportName, self.version)
        else:
            return "{0}:{1}:{2}:v{3}".format(self.name, self.kind, self.Script, self.version)

    def __repr__(self):
        """
        usual
        """
        if self.script is None:
            return "ModuleInstall({0}:{1}:import {2}:v{3})".format(
                self.name, self.kind, self.ImportName, self.version)
        else:
            return "{0}:{1}:{2}:v{3}".format(self.name, self.kind, self.Script, self.version)

    @property
    def ImportName(self):
        """
        return the import name
        """
        if self.mname is not None:
            return self.mname
        if self.name.startswith("python-"):
            return self.name[len("python-"):]
        else:
            return self.name

    def is_installed_local(self):
        """
        checks if a module is installed
        """
        if self.script is None:
            if self.skip_import:
                try:
                    return self.is_installed_local_cmd()
                except InstallError:
                    return False
            else:
                try:
                    if "." in self.ImportName:
                        raise ImportError(self.ImportName)
                    if sys.version_info[0] == 2:
                        r = importlib.import_module(self.ImportName)
                        return r
                    else:
                        r = importlib.util.find_spec(self.ImportName)
                        return r is not None
                    return r is not None
                except ImportError:
                    txt = "import {0}  # {1}".format(
                        self.ImportName, self.name)
                    try:
                        exec(txt)
                        return True
                    except Exception:
                        return False
        else:
            return os.path.exists(self.Script)

    def is_installed_local_cmd(self):
        """
        Test the module by running a command line.
        Does some others verifications for a specific modules such as scipy.

        .. versionadded:: 1.1
        """
        exe = get_python_program()
        if self.skip_import:
            cmd = exe + \
                ' -u -c "import pip # pip.main(["show", "{0}"])'.format(
                    self.name)
            out, err = run_cmd(cmd, fLOG=self.fLOG)
            if err or out is None:
                raise InstallError("Cannot find {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-K1:\n{3}".format(
                    self.name, cmd, out, err))
            if ("Name: " + self.name) not in out:
                raise InstallError("Cannot find {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-K3:\n{3}".format(
                    self.name, cmd, out, err))
            return True
        else:
            cmd = exe + \
                ' -u -c "import {0} # {1}"'.format(self.ImportName, self.name)
            out, err = run_cmd(cmd, fLOG=self.fLOG)
            if err:
                raise InstallError("cannot import module {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-K2:\n{3}".format(
                    self.ImportName, cmd, out, err))
            if self.name == "scipy":
                cmd = exe + ' -u -c "import scipy.sparse"'
                out, err = run_cmd(cmd, fLOG=self.fLOG)
                if err:
                    if sys.platform.startswith("win") and sys.version_info[:2] >= (3, 5) and "DLL" in err:
                        mes = ("scipy.sparse is failing, you should check that Visual Studio 2015 is " +
                               "installed\n{0}\nCMD:\n{1}\nOUT:\n{2}\nERR-M:\n{3}")
                        raise InstallError(mes.format(
                            self.ImportName, cmd, out, err))
                    raise InstallError("scipy.sparse is failing\n{0}\nCMD:\n{1}\nOUT:\n{2}\nERR-L:\n{3}".format(
                        self.ImportName, cmd, out, err))
            return True

    _page_cache_html2 = os.path.join(
        os.path.abspath(os.path.split(__file__)[0]), "page2.html")

    def get_exewheel_url_link2(self, file_save=None, wheel=False, source=None):
        """
        for windows, get the url of the setup using a webpage

        @param      file_save   for debug purposes
        @param      wheel       returns the wheel file or the exe file
        @param      source      source of the wheels (ex: ``2`` or ``http://...``)
        @return                 url, exe name

        .. versionchanged:: 1.1
            Parameter *source* was added.
        """
        if source is None or source == "2":
            source = ModuleInstall.exeLocationXd_Default
        source_page = source.rstrip("/") + "/index_modules_list.html"

        if "cached_page2" not in self.__dict__:
            page = ModuleInstall._page_cache_html2

            exi = os.path.exists(page)
            if exi:
                dt = get_file_modification_date(page)
                now = datetime.datetime.now()
                df = now - dt
                if df > datetime.timedelta(1):
                    exi = False

            if exi:
                text = read_page_wheel(page)
                self.cached_page2 = text
            else:
                text = get_page_wheel(source_page)
                save_page_wheel(page, text)
                self.cached_page2 = text

        page = self.cached_page2
        reg = re.compile('href=\\"(.*?)\\"')
        alls = reg.findall(page)
        if len(alls) == 0:
            keep = []
            for line in page.split("\n"):
                lline = line.lower()
                if self.name in lline or (self.mname and self.mname in lline):
                    keep.append(line)
            raise Exception(
                "module " +
                self.name + "\nexample:\n" + "\n".join(keep))

        version = python_version()
        plat = version[0] if version[0] == "win32" else version[1]
        if version[1] == '64bit' and version[0] == 'win32':
            plat = "amd64"
        cp = "-cp%d%d-" % sys.version_info[:2]
        py = "-py%d%d-" % sys.version_info[:2]
        pyn = "-py%d-" % sys.version_info[0]
        links = [_ for _ in alls if "/" +
                 self.name in _ and (pyn in _ or py in _ or cp in _) and (plat in _ or "-any" in _)]
        if len(links) == 0 and "-" in self.name:
            name_ = self.name.replace("-", "_")
            links = [_ for _ in alls if "/" +
                     name_ in _ and (pyn in _ or py in _ or cp in _) and (plat in _ or "-any" in _)]
        if len(links) == 0:
            # exception for guidata
            links = [_ for _ in alls if self.name in _ and "-py2.py3-" in _]
        if len(links) == 0:
            if file_save is not None:
                with open(file_save, "w", encoding="utf8") as f:
                    f.write(page)
            short_list = [_ for _ in alls if self.name in _]
            raise MissingWheelException("unable to find a single link for " +
                                        self.name + "\n" + "\n".join(short_list))

        # Last filter. Removes modules with a longer name.
        pref1 = self.name.lower() + "-"
        pref2 = self.name.lower().replace("-", "_") + "-"
        pref3 = self.name.lower().replace("_", "-") + "-"

        def filter_cond(name):
            name = name.split("/")[-1].lower()
            return name.startswith(pref1) or name.startswith(pref2) or name.startswith(pref3)

        links_ = [_ for _ in links if filter_cond(_)]
        if len(links_) == 0:
            prefs = "\npref1={0}\npref2={1}\npref3={2}".format(
                pref1, pref2, pref3)
            raise MissingWheelException("unable to find a single link for " +
                                        self.name + "\n" + "\n".join(links) + prefs)
        links = links_

        links = [(lu.split("/")[-1], lu) for lu in links]
        links0 = links

        if self.name == "numpy":
            links = [lu for lu in links if "unoptimized" not in lu[
                0].lower() and "vanilla" not in lu[0].lower()]

        if len(links) == 0:
            raise Exception("unable to find a single link for " +
                            self.name +
                            "\nEX:\n" +
                            "\n".join(str(_) for _ in links0))

        link = choose_most_recent(links)
        self.existing_version = self.extract_version(link[0])
        url, whl = link[1], link[0]
        if not whl.endswith(".whl"):
            whl += ".whl"
        return url, whl

    _page_cache_html = os.path.join(
        os.path.abspath(os.path.split(__file__)[0]), "page.html")

    def get_exewheel_url_link(self, file_save=None, wheel=False):
        """
        for windows, get the url of the setup using a webpage

        @param      file_save   for debug purposes
        @param      wheel       returns the wheel file or the exe file
        @return                 url, exe name
        """
        if "cached_page" not in self.__dict__:
            page = ModuleInstall._page_cache_html

            exi = os.path.exists(page)
            if exi:
                dt = get_file_modification_date(page)
                now = datetime.datetime.now()
                df = now - dt
                if df > datetime.timedelta(1):
                    exi = False

            if exi:
                text = read_page_wheel(page)
                self.cached_page = text
            else:
                text = get_page_wheel(ModuleInstall.exeLocation)
                save_page_wheel(page, text)
                self.cached_page = text

        page = self.cached_page
        alls = extract_all_links(page)
        if len(alls) == 0:
            keep = []
            for line in page.split("\n"):
                lline = line.lower()
                if self.name in lline or (self.mname and self.mname in lline):
                    keep.append(line)
            raise Exception(
                "module " +
                self.name + "\nexample:\n" + "\n".join(keep))

        version = python_version()
        plat = version[0] if version[0] == "win32" else version[1]
        if version[1] == '64bit' and version[0] == 'win32':
            plat = "amd64"
        links = list(enumerate_links_module(
            self.name, alls, sys.version_info, plat))
        if len(links) == 0:
            if file_save is not None:
                with open(file_save, "w", encoding="utf8") as f:
                    f.write(page)
            raise MissingWheelException(
                "Unable to find a single link for " + self.name)
        nbnone = [lu for lu in links if lu[2] is None]
        if len(nbnone) * 2 > len(links):
            raise WrongWheelException("Unable to find any version in\n{0}".format(
                "\n".join(str(_) for _ in links)))
        links0 = links

        if self.name == "numpy":
            links = [lu for lu in links if "unoptimized" not in lu[
                0].lower() and "vanilla" not in lu[0].lower()]

        if len(links) == 0:
            raise Exception("unable to find a single link for " +
                            self.name +
                            "\nEX:\n" +
                            "\n".join(str(_) for _ in links0))

        link = choose_most_recent(links)
        self.existing_version = self.extract_version(link[0])
        if link[2] is None:
            raise WrongWheelException("Unable to find a proper link in {0}\n{1}".format(
                link, "\n".join(str(_) for _ in links)))
        url, whl = ModuleInstall.exeLocation + link[2], link[0]
        if not whl.endswith(".whl"):
            whl += ".whl"
        return url, whl

    def unzipfiles(self, zipf, whereTo):
        """
        unzip files from a zip archive

        @param      zipf        archive
        @param      whereTo     destination folder
        @return                 list of unzipped files
        """
        return unzip_files(zipf, whereTo, self.fLOG)

    def extract_version(self, name):
        """
        extract the version from a filename

        @param      name        filename
        @return                 verions (str)
        """
        res = None
        for i, regve in enumerate(regex_wheel_versions):
            reg = re.compile(regve)
            res = reg.search(name)
            if res is not None:
                if i == 6:
                    return ".".join(res.groups())
                else:
                    return res.groups()[0]
        raise MissingVersionWheelException(
            "Unable to extract version number from '{0}'\nREGEX\n{1}".format(name, "\n".join(regex_wheel_versions)))

    def download(self, temp_folder=".", force=False, unzipFile=True,
                 file_save=None, deps=False, source=None):
        """
        Downloads the module without installation.

        @param      temp_folder     destination
        @param      force           force the installation even if already installed
        @param      unzipFile       if it can be unzipped, it will be (for github, mostly)
        @param      file_save       for debug purposes, do not change it unless you know what you are doing
        @param      deps            download the dependencies too (only available for pip)
        @param      source          overwrite source of the download, only for wheel packages,
                                    see @see me get_exewheel_url_link2
        @return                     downloaded files

        *deps* is overwritten by *self.deps* if not None
        If *source* is None, it is overwritten by *self.source*.
        """
        disable_options = {}
        if source is None:
            source = self.source
        kind = self.kind

        deps = deps if self.deps is None else self.deps

        if kind == "pip" or self.pipgit:
            # see https://pip.pypa.io/en/latest/reference/pip_install.html
            # we use pip install <package> --download=temp_folder
            pp = get_pip_program()
            if self.pipgit:
                br = "@" + \
                    self.branch if self.branch not in (None, "master") else ""
                link = ModuleInstall.github_pattern_git.format(
                    self.name, self.gitrepo, br)
                cmd = pp + ' download git+' + link
            else:
                cmd = pp + ' download {0}'.format(self.name)
            if self.version is not None:
                cmd += "=={0}".format(self.version)
            if " " in temp_folder:
                raise FileNotFoundError(
                    "no space allowed in folders: [" + temp_folder + "]")
            if deps:
                cmd += ' --dest={0}'.format(temp_folder)
            else:
                cmd += ' --dest={0} --no-deps'.format(temp_folder)

            if self.index_url is not None:
                slash = '' if self.index_url.endswith('/') else '/'
                cmd += ' --no-cache-dir --index={0}{1}simple/'.format(
                    self.index_url, slash)
                parsed_uri = urlsplit(self.index_url)
                cmd += " --trusted-host " + parsed_uri.hostname
            if self.pip_options is not None:
                diff = [_ for _ in self.pip_options if _ not in disable_options]
                cmd += " " + " ".join(diff)

            out, err = run_cmd(
                cmd, wait=True, fLOG=self.fLOG)
            if "Successfully downloaded" not in out:
                raise DownloadError(
                    "unable to download with pip " +
                    str(self) + "\nCMD:\n" +
                    cmd + "\nOUT:\n" +
                    out + "\nERR-N:\n" +
                    err)
            lines = out.split("\n")
            for line in lines:
                if line.strip().startswith("Saved "):
                    return line.split("Saved")[-1].strip()
                elif line.strip().startswith("File was already downloaded"):
                    return line.split("File was already downloaded")[-1].strip()
            raise FileNotFoundError(
                "unable to find downloaded file " +
                str(self) + "\nCMD:\n" +
                cmd + "\nOUT:\n" +
                out + "\nERR-O:\n" +
                err)

        if kind in ("wheel", "wheel2"):
            if source is not None:
                kind = "wheel2"
            ver = python_version()
            if ver[0] != "win32":
                # nothing to download, you should use pip
                return None
            else:
                if hasattr(self, "overwrite") and self.overwrite is not None:
                    over = self.overwrite.format(*sys.version_info[0:2])
                    url, whl = over, over.split("/")[-1]
                    self.existing_version = "{0}{1}".format(
                        *sys.version_info[0:2])
                elif kind == "wheel":
                    url, whl = self.get_exewheel_url_link(
                        file_save=file_save, wheel=True)
                else:
                    url, whl = self.get_exewheel_url_link2(
                        file_save=file_save, wheel=True, source=source)
                whlname = os.path.join(temp_folder, whl)

                exi = os.path.exists(whlname)
                if force or not exi:

                    self.fLOG("[pymy] downloading", whl)
                    # self.fLOG("url", url)
                    if self.existing_version is None:
                        self.existing_version = self.extract_version(whl)
                    req = urllib_request.Request(
                        url, headers={
                            'User-agent': default_user_agent})
                    try:
                        u = urllib_request.urlopen(req)
                        text = u.read()
                        u.close()
                    except urllib_error.HTTPError as e:
                        raise DownloadError("unable to download {} from {}".format(
                            os.path.split(whlname)[-1], url)) from e

                    if not os.path.exists(temp_folder):
                        os.makedirs(temp_folder)

                    if len(text) <= 1000:
                        raise WrongWheelException("Size of downloaded wheel is too small: {0}\nurl={1}\nagent={2}".format(
                            len(text), url, default_user_agent))

                    self.fLOG("[pymy] writing", whl)
                    with open(whlname, "wb") as f:
                        f.write(text)

                return whlname

        elif kind == "github" and not self.pipgit:
            outfile = os.path.join(temp_folder, self.name + ".zip")
            if force or not os.path.exists(outfile):
                zipurl = ModuleInstall.github_pattern_zip.format(
                    self.name, self.gitrepo, self.branch)
                self.fLOG("[pymy] downloading", zipurl)
                try:
                    req = urllib_request.Request(
                        zipurl, headers={
                            'User-agent': default_user_agent})
                    u = urllib_request.urlopen(req)
                    text = u.read()
                    u.close()
                except urllib_error.HTTPError as e:
                    raise Exception(
                        "unable to get archive from: " +
                        zipurl) from e

                if not os.path.exists(temp_folder):
                    os.makedirs(temp_folder)
                u = open(outfile, "wb")
                u.write(text)
                u.close()

            if unzipFile:
                self.fLOG("[pymy] unzipping ", outfile)
                files = self.unzipfiles(outfile, temp_folder)
                return files
            else:
                return outfile

        if kind in ("exe", "exe2"):
            if source is not None:
                kind = "exe2"
            ver = python_version()
            if ver[0] != "win32":
                raise Exception(
                    "this option is not available on other systems than Windows, version={0}".format(ver))
            url, exe = self.get_exewheel_url_link(
                file_save=file_save) if kind == "exe" else self.get_exewheel_url_link2(
                file_save=file_save, source=source)

            self.fLOG("[pymy] downloading", exe)
            req = urllib_request.Request(
                url, headers={
                    'User-agent': default_user_agent})
            u = urllib_request.urlopen(req)
            text = u.read()
            u.close()

            if not os.path.exists(temp_folder):
                os.makedirs(temp_folder)

            exename = os.path.join(temp_folder, exe)
            self.fLOG("[pymy] writing", exe)
            with open(exename, "wb") as f:
                f.write(text)
            return exename

        raise ImportError(
            "unknown kind: {0} for module {1}".format(
                kind,
                self.name))

    def get_pypi_version(self, url='https://pypi.python.org/pypi'):
        """
        returns the version of a package on pypi

        @param      url     pipy server
        @return             version

        See also `installing_python_packages_programatically.py <https://gist.github.com/rwilcox/755524>`_,
        `pkgtools.pypi: PyPI interface <http://pkgtools.readthedocs.org/en/latest/pypi.html>`_.
        """
        if url == 'https://pypi.python.org/pypi':
            # we use a function which caches the result
            return get_pypi_version(self.name)
        else:
            pypi = xmlrpc_client.ServerProxy(url)
            available = pypi.package_releases(self.name)
            if available is None or len(available) == 0:
                available = pypi.package_releases(self.name.capitalize())
            if (available is None or len(available) == 0) and self.mname is not None:
                available = pypi.package_releases(self.mname)

            if available is None:
                raise MissingPackageOnPyPiException(
                    "; ".join([self.name, self.name.capitalize(), self.mname]))

            return available[0]

    def get_pypi_numeric_version(self):
        """
        returns the version of a package in pypi

        @return     tuple
        """
        vers = self.get_pypi_version()
        if vers is None:
            return None
        if isinstance(vers, list):
            v = self.get_pypi_version()
            raise TypeError("unexpected type: {0} -- {1}".format(vers, v))
        return numeric_version(vers)

    def get_installed_version(self):
        """
        return the version of the installed package

        @return         version
        """
        vers = get_module_version(None)
        if self.name in vers:
            return vers[self.name]
        cap = self.name.capitalize()
        if cap in vers:
            return vers[cap]
        cap = self.name.lower()
        if cap in vers:
            return vers[cap]
        cap = self.name.replace("-", "_")
        if cap in vers:
            return vers[cap]
        cap = self.name.replace("_", "-")
        if cap in vers:
            return vers[cap]
        cap = self.name.lower().replace("_", "-")
        if cap in vers:
            return vers[cap]
        if self.mname is not None:
            if self.mname in vers:
                return vers[self.mname]
            cap = self.mname.lower()
            if cap in vers:
                return vers[cap]
        return None

    def get_installed_metadata(self):
        """
        return the metadata of the installed package

        @return         dictionary
        """
        r = get_module_metadata(self.name)
        if r is None:
            return get_module_metadata(self.mname)
        else:
            return r

    def get_installed_license(self):
        """
        return the license of the installed package

        @return         string
        """
        meta = self.get_installed_metadata()
        if meta is None or len(meta) == 0:
            res = None
        else:
            res = None
            for k, v in meta.items():
                if k.lower() == "license":
                    res = v
                    break
        adm = {None, "", "UNKNOWN"}
        if res is not None:
            if isinstance(res, list):
                res = [_ for _ in res if _ and _ not in adm]
                if len(res) > 0:
                    res = res[0]
                else:
                    res = None
        if res in adm:
            res = missing_module_licenses.get(self.name, None)
        return res

    def get_installed_classifier(self):
        """
        return the classifier of the installed package

        @return         string
        """
        meta = self.get_installed_metadata()
        if meta is None:
            return None
        for k, v in meta.items():
            if k.lower() == "classifier":
                return v
        return None

    def is_installed_version(self):
        """
        tells if a module is installed

        @return boolean
        """
        return self.get_installed_version() is not None

    def get_installed_numeric_version(self):
        """
        returns the version as number (not string)

        @return     tuple
        """
        vers = self.get_installed_version()
        if vers is None:
            return None
        return numeric_version(vers)

    def has_update(self):
        """
        tells if the package has a newer version on pipy

        @return boolean
        """
        if ModuleInstall.is_annoying(self.name):
            return False
        vers = self.get_installed_numeric_version()
        if self.version is None:
            pypi = self.get_pypi_numeric_version()
            return compare_version(pypi, vers) > 0
        else:
            num = numeric_version(self.version)
            return compare_version(num, vers) > 0

    def _check_installation(self):
        """
        some modules uninstall and install modules with another version number,
        we try to track that
        """
        try:
            import numpy
            if compare_version(numpy.__version__, "1.10") < 0:
                raise InstallError(
                    "numpy does not have a goof version number, it should be >= 1.10 not {0}".format(numpy.__version__))
        except ImportError:
            # not installed
            pass
        return True

    def install(self, force_kind=None, force=False, temp_folder=".",
                log=False, options=None, deps=False, source=None,
                custom=None, post=None, out_streams=None):
        """
        Installs the package.

        @param      force_kind      overwrite self.kind
        @param      force           force the installation even if already installed
        @param      temp_folder     folder where to download the setup
        @param      log             display logs or not
        @param      options         other options to add to the command line (see below) in a list
        @param      deps            install the dependencies too (only available for pip)
        @param      source          overwrite the source of the wheels,
                                    see @see me get_exewheel_url_link2
        @param      custom          overwrite parameters in ``self.custom``
        @param      post            instructions post installation (see the cnostructor for more help)
        @param      out_streams     if it is a list, the function will add standard outputs
        @return                     boolean

        The options mentioned in parameter ``options``
        are described here: `pip install <http://www.pip-installer.org/en/latest/usage.html>`_
        or `setup.py options <http://docs.python.org/3.4/install/>`_ if you
        installing a module from github.

        .. versionchanged:: 1.0
            *deps* is overwritten by *self.deps* if not None

        .. versionchanged:: 1.1
            On Anaconda (based on function @see fn is_conda_distribution), we try *conda* first
            before switching to the regular way if it did not work.
            Exception were changed from ``Exception`` to ``InstallError``.
            Parameter *source* was added, if None, it is overwritten by *self.source*.
            Parameter *custom* was added, it works the same as *source*.
            Parameter *post* was added.
            Parameter *out_streas* added.
        """
        if source is None:
            source = self.source
        if post is None:
            post = self.post_installation
        if not force and force_kind is None and is_conda_distribution():
            try:
                return self.install(force_kind="conda", force=True, temp_folder=temp_folder,
                                    log=log, options=options, deps=deps)
            except InstallError as e:
                warnings.warn(str(e))
                # we try the regular way now

        if not force and self.is_installed_version():
            return True

        deps = deps if self.deps is None else self.deps

        if options is None:
            options = self.pip_options
        if options is None:
            options = []

        kind = force_kind if force_kind is not None else self.kind
        add = (" with " + kind) if kind != self.kind else ""
        self.fLOG("[pymy] installation of " + str(self) + add)
        ret = None
        custom = custom or self.custom

        if kind == "pip" or self.pipgit:
            if custom is not None:
                raise NotImplementedError(
                    "custom must be None not '{0}' when kind is '{1}'".format(custom, kind))
            pp = get_pip_program()
            if self.pipgit:
                br = "@" + \
                    self.branch if self.branch not in (None, "master") else ""
                link = ModuleInstall.github_pattern_git.format(
                    self.name, self.gitrepo, br)
                cmd = pp + ' install git+' + link
            else:
                cmd = pp + " install {0}".format(self.name)
            if self.version is not None:
                cmd += "=={0}".format(self.version)
            if len(options) > 0:
                cmd += " " + " ".join(options)

            if not deps:
                cmd += ' --no-deps'

            if self.index_url is not None:
                slash = '' if self.index_url.endswith('/') else '/'
                cmd += ' --no-cache-dir --index={0}{1}simple/'.format(
                    self.index_url, slash)
            else:
                cmd += " --cache-dir={0}".format(temp_folder)

            if self.name == "kivy-garden":
                memo = sys.argv
                sys.argv = []
            out, err = run_cmd(
                cmd, wait=True, fLOG=self.fLOG)
            out = _filter_pip_out(out)
            if out_streams is not None:
                out_streams.append((cmd, out, err))
            if self.name == "kivy-garden":
                sys.argv = memo

            success2 = "Requirement already up-to-date: " + self.name
            uptodate = success2.replace("-", "_") in out.replace("-", "_")

            if "No distributions matching the version" in out:
                mes = "(1) unable to install with pip {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-P:\n{3}".format(
                    str(self), cmd, out, err)
                raise InstallError(mes)
            if "Testing of typecheck-decorator passed without failure." in out:
                ret = True
            elif "Successfully installed" not in out and not uptodate:
                if "error: Unable to find vcvarsall.bat" in out:
                    url = "http://www.xavierdupre.fr/blog/2013-07-07_nojs.html"
                    mes = "(2) unable to install with pip {0}\nread:\n{1}\nCMD:\n{2}\nOUT:\n{3}\nERR-P:\n{4}".format(
                        str(self), url, cmd, out, err)
                    raise InstallError(mes)
                if "Requirement already satisfied" not in out and not uptodate:
                    mes = "(3) unable to install with pip {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-Q:\n{3}".format(
                        str(self), cmd, out, err)
                    raise InstallError(mes)
            else:
                ret = not uptodate

        elif kind == "conda":
            if custom is not None:
                raise NotImplementedError(
                    "custom must be None not '{0}' when kind is '{1}'".format(custom, kind))
            if "--upgrade" in options:
                options = [_ for _ in options if _ != "--upgrade"]
                command = "update"
            else:
                command = "install"
            pp = get_conda_program()
            cmd = pp + " {1} {0}".format(self.name, command)
            if self.version is not None:
                cmd += "=={0}".format(self.version)
            if len(options) > 0:
                cmd += " " + " ".join(options)

            cmd += " --no-pin --yes --quiet"
            if not deps:
                cmd += ' --no-deps'

            out, err = run_cmd(
                cmd, wait=True, fLOG=self.fLOG)
            if out_streams is not None:
                out_streams.append((cmd, out, err))
            if "No distributions matching the version" in out or \
               "No packages found in current linux" in out:
                mes = "(4) unable to install with conda {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-R:\n{3}".format(
                    str(self), cmd, out, err)
                raise InstallError(mes)
            if "Testing of typecheck-decorator passed without failure." in out:
                ret = True
            elif "Successfully installed" not in out:
                if "error: Unable to find vcvarsall.bat" in out:
                    url = "http://www.xavierdupre.fr/blog/2013-07-07_nojs.html"
                    mes = "(5) unable to install with conda {0}\nread:\n{1}\nCMD:\n{2}\nOUT:\n{3}\nERR-S:\n{4}".format(
                        str(self), url, cmd, out, err)
                    raise InstallError(mes)
                if "Requirement already satisfied" not in out:
                    mes = "(6) unable to install with conda {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-T:\n{3}".format(
                        str(self), cmd, out, err)
                    raise InstallError(mes)
            else:
                ret = True

        elif kind in ("wheel", "wheel2"):
            if custom is not None:
                raise NotImplementedError(
                    "custom must be None not '{0}' when kind is '{1}'".format(custom, kind))
            ver = python_version()
            if ver[0] != "win32":
                ret = self.install("pip")
                whlname = self.name
                ret = True
            else:
                whlname = self.download(
                    temp_folder=temp_folder,
                    force=force,
                    unzipFile=True,
                    source=source)
                vers = self.get_installed_numeric_version()
                ret = True
                if vers is not None:
                    whlvers = numeric_version(get_wheel_version(whlname))
                    if compare_version(vers, whlvers) >= 0:
                        self.fLOG("[pymy] skipping, no newer version {0} <= {1}: whl= {2}".format(
                            whlvers, vers, whlname))
                        ret = False
            if ret:
                self.fLOG("[pymy] installing", os.path.split(whlname)[-1])

                pip = get_pip_program()
                cmd = pip + " install {0}".format(whlname)
                if self.version is not None:
                    cmd += "=={0}".format(self.version)
                if len(options) > 0:
                    opts = list(options)
                    if len(opts):
                        cmd += " " + " ".join(opts)
                if not deps:
                    cmd += ' --no-deps'

                out, err = run_cmd(
                    cmd, wait=True, fLOG=self.fLOG)
                if out_streams is not None:
                    out_streams.append((cmd, out, err))
                if "No distributions matching the version" in out:
                    mes = "(7) unable to install with wheel {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-U:\n{3}".format(
                        str(self), cmd, out, err)
                    raise InstallError(mes)
                if "Testing of typecheck-decorator passed without failure." in out:
                    ret = True
                elif "Successfully installed" not in out:
                    if "error: Unable to find vcvarsall.bat" in out:
                        url = "http://www.xavierdupre.fr/blog/2013-07-07_nojs.html"
                        mes = "(8) unable to install with wheel {0}\nread:\n{1}\nCMD:\n{2}\nOUT:\n{3}\nERR-V:\n{4}".format(
                            str(self), url, cmd, out, err)
                        raise InstallError(mes)
                    if "Requirement already satisfied" not in out and "Requirement already up-to-date" not in out:
                        mes = "(9) unable to install with wheel {0}\nCMD:\n{1}\nOUT:\n{2}\nERR-W:\n{3}".format(
                            str(self), cmd, out, err)
                        raise InstallError(mes)
                else:
                    ret = True

        elif kind == "github" and not self.pipgit:
            # the following code requires admin rights
            # if python_version()[0].startswith("win") and kind == "git" and not os.path.exists(ModuleInstall.gitexe) :
            #    raise FileNotFoundError("you need to install github first: see http://windows.github.com/")
            # if python_version()[0].startswith("win"):
            #    os.chdir(os.path.join(ModuleInstall.gitexe,"bin"))
            # cmd = pip + " install -e
            # git://github.com/{0}/{1}-python.git#egg={1}".format(self.gitrepo,
            # self.name)

            files = self.download(temp_folder=temp_folder,
                                  force=force, unzipFile=True)
            setu = [_ for _ in files if _.endswith("setup.py")]
            if len(setu) == 0:
                raise Exception(
                    "unable to find setup.py for module " +
                    self.name)
            if len(setu) > 1:
                setu = [(len(_), _) for _ in setu]
                setu.sort()
                if setu[0][0] == setu[1][0]:
                    raise InstallError(
                        "(10) more than one setup.py for module " +
                        self.name +
                        "\n" +
                        "\n".join(
                            str(_) for _ in setu))
                self.fLOG(
                    "[pymy] warning: more than one setup: " + str(setu))
                setu = [setu[0][1]]
            setu = os.path.abspath(setu[0])

            self.fLOG("[pymy] install ", setu[0])
            cwd = os.getcwd()
            os.chdir(os.path.split(setu)[0])
            if custom is None:
                custom = ["install"]
            cmds = []
            for command in custom:
                cmd1 = "{0} setup.py {1}".format(
                    sys.executable.replace("pythonw.exe", "python.exe"),
                    command)
                cmds.append(cmd1)

            def enumerate_filtered_option(options):
                for o in options:
                    if o not in ('--no-deps', '--upgrade'):
                        yield o

            filter_options = list(enumerate_filtered_option(options))
            if len(filter_options) > 0:
                cmds[-1] += " " + " ".join(filter_options)

            if deps:
                # it will not work
                # cmd += ' --no-deps'
                pass

            outs = ""
            errs = ""
            for cmd in cmds:
                out, err = run_cmd(
                    cmd, wait=True, fLOG=self.fLOG)
                if out_streams is not None:
                    out_streams.append((cmd, out, err))
                if len(outs) > 0:
                    outs += "\n"
                if len(errs) > 0:
                    errs += "\n"
                outs += out
                errs += err
            os.chdir(cwd)

            out, err = outs, errs
            if "Successfully installed" not in out and "install  C" not in out:
                if "Finished processing dependencies" not in out:
                    raise InstallError(
                        "(11) unable to install with github " +
                        str(self) +
                        "\nOUT:\n" +
                        out +
                        "\nERR-X:\n" +
                        err)
                self.fLOG(
                    "warning: ``Successfully installed`` or ``install  C`` not found")
                if "Permission denied" in out:
                    raise PermissionError(" ".join(["(12) unable to install with github", str(self), "\n--CMD--:\n",
                                                    "\n".join(cmds), "\n--OUT--\n", out, "\n--ERR-Y--\n", err]))
            ret = True

        elif kind == "exe":
            if custom is not None:
                raise NotImplementedError(
                    "custom must be None not '{0}' when kind is '{1}'".format(custom, kind))
            ver = python_version()
            if ver[0] != "win32":
                ret = self.install("pip")
            else:
                exename = self.download(
                    temp_folder=temp_folder,
                    force=force,
                    unzipFile=True,
                    source=source)
                self.fLOG("[pymy] executing", os.path.split(exename)[-1])
                out, err = run_cmd(
                    exename + " /s /qn /SILENT", wait=True, fLOG=self.fLOG)
                if out_streams is not None:
                    out_streams.append((cmd, out, err))
                ret = len(err) == 0

        elif kind == "exe2":
            if custom is not None:
                raise NotImplementedError(
                    "custom must be None not '{0}' when kind is '{1}'".format(custom, kind))
            ver = python_version()
            if ver[0] != "win32":
                ret = self.install("pip")
            else:
                exename = self.download(
                    temp_folder=temp_folder,
                    force=force,
                    unzipFile=True,
                    source=source)
                self.fLOG("[pymy] executing", os.path.split(exename)[-1])
                out, err = run_cmd(
                    exename + " /s /qn", wait=True, fLOG=self.fLOG)
                if out_streams is not None:
                    out_streams.append((cmd, out, err))
                ret = len(err) == 0
        else:
            raise ImportError(
                "unknown kind: {0} for module {1}".format(
                    kind,
                    self.name))

        if ret is not None and ret:
            self._check_installation()

        # at this stage, there is a bug, for some executable, the execution
        # takes longer than expected
        # if not self.is_installed_version() :
        #    raise Exception("unable to install module: {0}, str:{1}".format(self.name, self))

        if ret is not None and ret and self.script is not None:
            if sys.platform.startswith("win"):
                # here, we have to wait until the script is installed
                ti = 0
                while not self.is_installed_local():
                    time.sleep(0.5)
                    ti += 0.5
                    if ti > 60:
                        self.fLOG(
                            "wait for the end of execution of ",
                            self.name)
                    ti = 0

                # we add set path=%path%;PYTHONPATH
                with open(self.Script, "r") as f:
                    full = f.read()

                exe = os.path.split(sys.executable)[0]
                cmd = "set path=%path%;{0}".format(exe)
                if cmd not in full:
                    self.fLOG("[pymy] add {0} to {1}".format(cmd, self.Script))
                    with open(self.Script, "w") as f:
                        if full.startswith("@echo off"):
                            f.write(
                                full.replace(
                                    "@echo off",
                                    "@echo off\n{0}".format(cmd)))
                        else:
                            f.write(cmd)
                            f.write("\n")
                            f.write(full)

        if ret is not None and post is not None:
            self.fLOG("[pymy] _ run_post_installation [begin]", post)
            ret = self.run_post_installation(post)
            self.fLOG("[pymy] _ run_post_installation [end]")

        if ret is not None and ret:
            # we check the module was properly installed
            if not self.is_installed_local_cmd():
                raise InstallError(
                    "** unable to install module {0}, unable to import it".format(self.name))

        return ret

    def run_post_installation(self, post):
        """
        Run instructions post installation

        @param      post        dictionary, instructions post installation
        @return                 boolean

        Example:

        ::

            post = dict(
                cmd_python="Scripts\\\\pywin32_postinstall.py -install")

        .. versionadded:: 1.1
        """
        if not isinstance(post, dict):
            raise TypeError("expecting a dictionary")
        for k, v in post.items():
            if k == "cmd_python":
                # processed just above
                continue
            if k == "pre_cmd":
                if v == "module_install_preprocess":
                    self.fLOG("[pymy] _ module_install_preprocess [begin]")
                    self.module_install_preprocess(post)
                    self.fLOG("[pymy] _ module_install_preprocess [end]")
                else:
                    raise ValueError(
                        "Unable to interpret value '{0}'.".format(v))
            else:
                raise KeyError("Unable to interpret command '{0}'".format(k))
        for k, v in post.items():
            if k == "cmd_python":
                exe = os.path.abspath(sys.executable)
                cmd = '"{0}" {1}'.format(exe, v.format(os.path.dirname(exe)))
                out, err = run_cmd(cmd, wait=True, fLOG=self.fLOG)
                if err is not None and len(err) > 0:
                    raise InstallError(
                        "Post installation script failed.\nCMD\n{0}\nOUT\n{1}\nERR-Z\n{2}".format(cmd, out, err))
                self.fLOG("[pymy] OUT:\n{0}".format(out))
            elif k == "pre_cmd":
                # processed just above
                continue
            else:
                raise KeyError("Unable to interpret command '{0}'".format(k))
        return True

    def update(self, force_kind=None, force=False, temp_folder=".",
               log=False, options=None, deps=False, source=None):
        """
        Updates the package if necessary, we use
        ``pip install <module_name> --upgrade --no-deps``,

        @param      force_kind      overwrite self.kind
        @param      force           force the installation even if not need to update
        @param      temp_folder     folder where to download the setup
        @param      log             display logs or not
        @param      options         others options to add to the command line (see below) an a list
        @param      deps            download the dependencies too (only available for pip)
        @param      source          overwrite the source of the wheel, see @see me get_exewheel_url_link2
        @return                     boolean

        The options mentioned in parameter ``options``
        are described here: `pip install <http://www.pip-installer.org/en/latest/usage.html>`_
        or `setup.py options <http://docs.python.org/3.4/install/>`_ if you
        installing a module from github.

        .. versionchanged:: 1.1
            Parameter *source* was added, if None, it is overwritten by *self.source*.
        """
        if source is None:
            source = self.source
        if ModuleInstall.is_annoying(self.name):
            return False

        if not self.is_installed_version():
            raise MissingInstalledPackageException(self.name)

        if not force and not self.has_update():
            return True

        self.fLOG("[pymy] update of ", self)

        options = [] if options is None else list(options)
        for opt in ["--upgrade", "--no-deps"]:
            if opt not in options:
                if not deps or opt == "--no-deps":
                    options.append(opt)

        res = self.install(force_kind=force_kind, force=True,
                           temp_folder=temp_folder, log=log, options=options, source=source)
        return res

    def module_install_preprocess(self, post):
        """
        Run some preprocessing for specific modules

        @param      post        dictionary
        """
        if self.name == "pywin32":
            self.fLOG("[pymy] _ module_install_preprocess", self.name)
            if "cmd_python" not in post:
                raise KeyError("Key 'cmd_python' is not in post.")
            cmd_python = post["cmd_python"]
            name = cmd_python.split()[0].format(
                os.path.dirname(sys.executable))
            self.fLOG("[pymy] _ opening", name)
            if not os.path.exists(name):
                raise FileNotFoundError(name)
            with open(name, "r") as f:
                content = f.read()
            repby = "os.path.exists(dest)"
            if repby not in content:
                self.fLOG("[pymy] _ Preprocess '{0}'".format(name))
                rep = "def CopyTo(desc, src, dest):"
                content = content.replace(
                    rep, "{0}\n    if {1}: return".format(rep, repby))
                with open(name, "w") as f:
                    f.write(content)
                self.fLOG("[pymy] _ Done.")
            else:
                self.fLOG(
                    "Skip preprocess '{0}' because '{1}' was found.".format(name, repby))
        else:
            raise ValueError(
                "No preprocessing for module '{0}'.".format(self.name))



"""
@file
@brief Various functions to install `Julia <http://www.julialang.org/>`_.
"""
from __future__ import print_function
import sys
import re
import os

from ..installhelper.install_cmd_helper import run_cmd
from .install_custom import download_page, download_file


def get_julia_version():
    """
    returns the version of installed R,
    we only focus on the x64 version

    @return     tuple (bin, version), None if R is not installed
    """
    if sys.platform.startswith("win"):
        path = "{0}\\Julia".format(os.environ["ProgramFiles"])
        if os.path.exists(path):
            vers = os.listdir(path)
            vers.sort()
            if len(vers) == 0:
                return None
            vers = vers[-1]
            bin = os.path.join(path, vers, "bin", "x64")
            if not os.path.exists(bin):
                return None
            return (bin, vers)
        return None
    else:
        raise NotImplementedError("not available on platform " + sys.platform)


def IsJuliaInstalled():
    """
    @return     True of False whether or not it was installed
    """
    r = get_julia_version()
    return r is not None


def install_julia(
        temp_folder=".", fLOG=print, install=True, force_download=False, version=None):
    """
    Install `R <http://www.r-project.org/>`_.
    It does not do it a second time if it is already installed.

    @param      temp_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      force_download  force the downloading of Julia
    @param      version         version to download (unused)
    @return                     temporary file
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    bb = IsJuliaInstalled()
    if bb and not force_download:
        return True

    link = "http://julialang.org/downloads/"
    page = download_page(link)
    if sys.platform.startswith("win"):
        reg = re.compile("href=\\\"(.*?win64[.]exe)\\\"")
        alls = reg.findall(page)
        if len(alls) == 0:
            raise Exception(
                "unable to find a link on a .exe file on page: " +
                page)

        url = alls[0]
        full = url.split("/")[-1]
        outfile = os.path.join(temp_folder, full)
        fLOG("[pymy] download ", url)
        local = download_file(url, outfile)
        if install and not bb:
            run_cmd("msiexec /i " + local, fLOG=fLOG, wait=True)
        return local
    else:
        raise NotImplementedError("not available on platform " + sys.platform)



# -*- coding: utf-8 -*-
"""
@file
@brief Defines different a set of usual modules for Python.
"""
import sys
from ..installhelper.module_install import ModuleInstall
from ..installhelper.install_cmd_helper import is_conda_distribution


def small_set():
    """
    list of modules to work with pandas, numpy, ipython, ...
    """
    mod = [
        # issues with this modules in a virtual environment raise a, b, c instead or raise (a, b, c)
        # ModuleInstall("futures", "pip", mname="concurrent.futures",
        # purpose="Backport of the concurrent.futures package from Python
        # 3.2"),
        ModuleInstall(
            "wincertstore", "pip", purpose="Python module to extract CA and CRL certs from Windows' cert store (ctypes based).") \
        if sys.platform.startswith("win") else None,
        ModuleInstall(
            "pyquicksetup", "pip", purpose="helpers for setup", usage="TEACH"),
        ModuleInstall(
            "virtualenv", "pip", purpose="creatre virtual environments") if not is_conda_distribution() else None,
        ModuleInstall(
            "six", "pip", purpose="helpers for python 2/3 conversion"),
        ModuleInstall("lxml", "wheel", purpose="xml parsers (C++)"),
        ModuleInstall('markupsafe', 'pip', purpose="parses mardown"),
        ModuleInstall("jinja2", "pip", purpose="templating"),
        ModuleInstall("Mako", "pip", mname="mako", purpose="templating"),
        ModuleInstall(
            "pygments", "pip", purpose="syntax highlighting package written in Python"),
        ModuleInstall(
            "pyparsing", "pip", purpose="alternative approach to creating and executing simple grammars"),
        ModuleInstall("idna", "pip", usage="WEB",
                      purpose="Internationalized Domain Names in Applications (IDNA)"),
        ModuleInstall(
            "python-dateutil", "pip", "dateutil", purpose="helpers to manipulate dates"),
        ModuleInstall(
            "webencodings", "pip", purpose="Character encoding aliases for legacy web content"),
        ModuleInstall("soupsieve", "pip",
                      purpose="Soup Sieve is a CSS selector library designed to be used with Beautiful Soup 4."),
        ModuleInstall("beautifulsoup4", "pip", mname="bs4",
                      purpose="Beautiful Soup sits atop an HTML or XML parser, providing Pythonic idioms for " +
                      "iterating, searching, and modifying the parse tree."),
        ModuleInstall(
            "certifi", "pip", purpose="Certifi is a carefully curated collection of Root Certificates for validating the trustworthiness " +
            "of SSL certificates while verifying the identity of TLS hosts."),
        ModuleInstall("requests", "pip", purpose="human interface for http"),
        ModuleInstall(
            "coverage", "pip", purpose="measure the coverage of unit tests"),
        ModuleInstall("urllib3", "pip",
                      purpose="urllib2 extension"),
        ModuleInstall(
            "chardet", "pip", usage="WEB",
            purpose="Universal encoding detector."),
        ModuleInstall(
            "codecov", "pip", purpose="submit coverage report to codecov"),
        ModuleInstall("nose", "pip", purpose="run unit tests"),
        ModuleInstall(
            "pytz", "pip", purpose="World timezone definitions, modern and historical"),
        ModuleInstall("pyreadline", "pip",
                      purpose="python implementation of GNU readline functionality"),
        ModuleInstall(
            "pipdeptree", "pip", purpose="displays module dependencies as a tree"),
        ModuleInstall("jdcal", "pip",
                      purpose="Julian dates from proleptic Gregorian and Julian calendars."),
        ModuleInstall('et_xmlfile', "pip",
                      purpose="et_xmlfile is a low memory library for creating large XML files (for openpyxl)."),
        ModuleInstall("openpyxl", "pip",
                      purpose="reads/writes Excel files, version is 1.8.6 due to pandas which does not work with more recent verrsions yet"),
        ModuleInstall("xlrd", "pip", purpose="reads Excel files"),
        ModuleInstall("xlwt", "pip", purpose="writes Excel files"),
        ModuleInstall(
            'XlsxWriter', 'pip', mname='xlsxwriter', purpose="writes Excel files"),
        ModuleInstall(
            "tornado", "wheel", purpose="python server, IPython relies on it", usage="NETWORK"),
        ModuleInstall(
            "sockjs-tornado", "pip", mname="sockjs.tornado", usage="NETWORK",
            purpose="SockJS-tornado is a Python server side counterpart of SockJS-client browser library running " +
            "on top of Tornado framework."),
        ModuleInstall(
            "pyzmq", "wheel", mname="zmq", purpose="python librairies for Omz (pipy distributes the binaries)", usage="NETWORK"),
        ModuleInstall(
            "pycparser", "pip", purpose="pycparser is a complete parser of the C language, written in pure " +
            "Python using the PLY parsing library. It parses C code into " +
            "an AST and can serve as a front-end for C compilers or analysis tools."),
        ModuleInstall("numpy", "wheel",
                      purpose="matrix computation", usage="DATA/ML"),
        ModuleInstall("mkl_service", "wheel", mname="mkl",
                      purpose="This package exposes a few functions which are declared in mkl_service.h. The main purpose of the " +
                      "package is to allow the user to change the number of CPU's MKL is using at runtime.", usage="DATA/ML"),
        ModuleInstall("Cython", "wheel", mname="cython",
                      purpose="pseudo C++ in python"),
        ModuleInstall("cycler", "pip",
                      purpose="dependency for matplotlib", usage="VIZ"),
        ModuleInstall("kiwisolver", "wheel",
                      purpose="Kiwi is an efficient C++ implementation of the Cassowary constraint solving algorithm."),
        ModuleInstall("matplotlib", "wheel",
                      purpose="most used plotting library", usage="VIZ"),
        ModuleInstall("mpl_finance", "github", "matplotlib", usage="VIZ",
                      purpose="This module consists of code extracted from the deprecated matplotlib.finance " +
                      "module along with a few examples of usage."),
        ModuleInstall("pywin32", "wheel2",
                      mname="win32com", purpose="call Windows DLL",
                      post=dict(
                          pre_cmd="module_install_preprocess",
                          cmd_python="{0}\\Scripts\\pywin32_postinstall.py -install"),
                      usage="WINDOWS") if sys.platform.startswith("win") else None,
        ModuleInstall("pywin32-ctypes", "pip", mname="win32ctypes",
                      purpose="call Windows DLL", usage="WINDOWS") if sys.platform.startswith("win") else None,
        ModuleInstall("winshell", "pip", purpose="Windows shell functions",
                      usage="WINDOWS") if sys.platform.startswith("win") else None,
        ModuleInstall(
            "brewer2mpl", "pip", purpose="Connect colorbrewer2.org color maps to Python and matplotlib"),
        ModuleInstall(
            "scipy", "wheel", purpose="scientific computation, eigen values, linear algebra", usage="DATA/ML"),
        ModuleInstall(
            "patsy", "pip", purpose="A Python package for describing statistical models and for building design matrices.", usage="DATA/ML"),
        ModuleInstall(
            "decorator", "pip", purpose="Better living through Python with decorators"),
        ModuleInstall(
            "networkx", "pip", purpose="graph libraries, basic drawing", usage="VIZ"),
        ModuleInstall(
            "graphviz", "pip", purpose="wrapper for graphviz (most famous tool to draw graphs", usage="VIZ"),
        ModuleInstall("pyrsistent", "pip", purpose="Pyrsistent is a number of persistent collections (by some referred to as functional data "
                      "structures). Persistent in the sense that they are immutable."),
        ModuleInstall('attrs', 'pip',
                      purpose="Attributes without boilerplate."),
        ModuleInstall(
            "jsonschema", "pip", purpose="An implementation of JSON Schema validation for Python"),
        ModuleInstall(
            "mistune", "pip", purpose="The fastest markdown parser in pure Python with renderer features, inspired by marked."),
        ModuleInstall("wheel", "pip", purpose="handle wheels"),
        ModuleInstall(
            "alabaster", "pip", purpose="A configurable sidebar-enabled Sphinx theme", usage="SPHINX"),
        ModuleInstall("babel", "pip",  # version="1.3",
                      purpose="Internationalization utilities, version 2.0 has bugs", usage="SPHINX"),
        ModuleInstall(
            "colorama", "pip", purpose="Cross-platform colored terminal text.", usage="SPHINX"),
        ModuleInstall(
            "wcwidth", "pip", purpose="Measures number of Terminal column cells of wide-character codes", usage="JUPYTER"),
        ModuleInstall(
            "prompt_toolkit", "pip", purpose="Library for building powerful interactive command lines in Python", usage="JUPYTER"),
        ModuleInstall("docutils", "pip",
                      purpose="interpret RST format", usage="SPHINX"),
        ModuleInstall("packaging", "pip",
                      purpose="Core utilities for Python packages"),
        ModuleInstall("sphinxcontrib-applehelp", "pip", mname="sphinxcontrib.applehelp",
                      purpose="extension for Sphinx", usage="SPHINX"),
        ModuleInstall("sphinxcontrib-devhelp", "pip", mname="sphinxcontrib.devhelp",
                      purpose="extension for Sphinx", usage="SPHINX"),
        ModuleInstall("sphinxcontrib-htmlhelp", "pip", mname="sphinxcontrib.htmlhelp",
                      purpose="extension for Sphinx", usage="SPHINX"),
        ModuleInstall("sphinxcontrib-jsmath", "pip", mname="sphinxcontrib.jsmath",
                      purpose="extension for Sphinx", usage="SPHINX"),
        ModuleInstall("sphinxcontrib-qthelp", "pip", mname="sphinxcontrib.qthelp",
                      purpose="extension for Sphinx", usage="SPHINX"),
        ModuleInstall("sphinxcontrib-serializinghtml", "pip", mname="sphinxcontrib.serializinghtml",
                      purpose="extension for Sphinx", usage="SPHINX"),
        ModuleInstall("sphinxcontrib-websupport", "pip", mname="sphinxcontrib.websupport",
                      purpose="extension for Sphinx", usage="SPHINX"),
        ModuleInstall(
            "sphinx", "pip", purpose="documentation generation based on RST", usage="SPHINX"),
        ModuleInstall(
            'imagesize', 'pip', usage="SPHINX", purpose="Getting image size from png/jpeg/jpeg2000/gif file"),
        # ModuleInstall(
        #     'sphinxcontrib-images', 'github', "sdpython", mname='sphinxcontrib.images',
        #      usage="SPHINX", purpose="include images in Sphinx documentation"),
        ModuleInstall('pypiserver', 'pip',
                      purpose="run a local pypi server"),
        ModuleInstall(
            "untokenize", "pip", purpose="Transforms tokens into original source code (while preserving whitespace).", usage="STYLE"),
        ModuleInstall(
            "pycodestyle", "pip", purpose="Python style guide checker", usage="STYLE"),
        ModuleInstall(
            "pep8", "pip", purpose="official guidelines on Python style"),
        ModuleInstall("autopep8", "pip",
                      purpose="apply pep8 on a script", usage="STYLE"),
        ModuleInstall("docformatter", "pip",
                      purpose="Formats docstrings to follow PEP 257.", usage="STYLE"),
        ModuleInstall("unify", "pip",
                      purpose="Modifies strings to all use the same (single/double) quote where possible.", usage="STYLE"),
        ModuleInstall("mccabe", "pip", usage="STYLE",
                      purpose="This module provides a plugin for flake8, the Python code checker."),
        ModuleInstall("pyflakes", "pip",
                      purpose="verify pep8 on a script", usage="STYLE"),
        ModuleInstall("flake8", "pip",
                      purpose="verify pep8 on a script", usage="STYLE"),
        ModuleInstall(
            "bottleneck", "wheel", purpose="pandas depenencies, faster arrays", usage="DATA/ML"),
        ModuleInstall(
            "pandas", "wheel", purpose="manipulate table as SQL in memory", usage="DATA/ML"),
        ModuleInstall(
            "statsmodels", "wheel", purpose="statistical modelling, depends on scipy", usage="DATA/ML"),
        ModuleInstall(
            "enum34", "pip", purpose="for ggplot, Object-oriented filesystem paths") if sys.version_info[:2] < (3, 4) else None,
        ModuleInstall("fire", "pip",
                      purpose="Python Fire is a library for automatically generating command line "
                      "interfaces (CLIs) from absolutely any Python object."),
        ModuleInstall("pathlib2", "pip",
                      purpose="for ggplot, Object-oriented filesystem paths"),
        ModuleInstall("ggplot", "pip", purpose="ggplot graphics style"),
        ModuleInstall("plotnine", "pip",
                      purpose="A grammar of graphics for Python"),
        ModuleInstall(
            "requests-file", "pip", mname="requests_file", purpose="File transport adapter for Requests"),
        ModuleInstall(
            "requests-ftp", "pip", mname="requests_ftp", purpose="FTP Transport Adapter for Requests"),
        ModuleInstall(
            "pandas-datareader", "pip", mname="pandas_datareader",
            purpose="Up to date remote data access for pandas, works for multiple versions of pandas.", usage="DATA/ML"),
        ModuleInstall("cftime", "wheel",
                      purpose="Time-handling functionality from netcdf4-python."),
        ModuleInstall("netCDF4", "wheel",
                      purpose="xarray uses this module to save and read data (netCDF=Unidata network Common Data Form)"),
        ModuleInstall(
            "xarray", "pip", purpose="pandas like library for cubes (N-dimensional data)", usage="DATA/ML"),
        ModuleInstall(
            "threadpoolctl", "pip", purpose="Python helpers to limit the number of threads used in the "
            "threadpool-backed of common native libraries used for scientific computing and data science "
            "(e.g. BLAS and OpenMP)."),
        ModuleInstall(
            "joblib", "pip", purpose="distribute jobs, parallelization"),
        ModuleInstall(
            "scikit-learn", "wheel", mname="sklearn", purpose="machine learning", usage="DATA/ML"),
        # ipython
        ModuleInstall("pandocfilters", "pip",
                      purpose="Utilities for writing pandoc filters in python"),
        ModuleInstall("pandoc-attributes", "pip", mname="pandocattributes",
                      purpose="An Attribute class to be used with pandocfilters"),
        ModuleInstall("win_unicode_console", "pip",
                      "Enable Unicode input and display when running Python from Windows console."),
        ModuleInstall(
            "ipython_genutils", "pip", purpose="IPython utils (nbformat)", usage="JUPYTER"),
        ModuleInstall(
            "html5lib", "pip", purpose="pure-python library for parsing HTML"),
        ModuleInstall("bleach", "pip", usage="WEB",
                      purpose="An easy whitelist-based HTML-sanitizing tool."),
        ModuleInstall(
            "testpath", "pip", purpose="Test utilities for code working with files and commands"),
        ModuleInstall(
            "traitlets", "pip", purpose="IPython, dependency", usage="JUPYTER"),
        ModuleInstall(
            "pickleshare", "pip", purpose="IPython, dependency", usage="JUPYTER"),
        ModuleInstall(
            "simplegeneric", "pip", purpose="IPython, dependency", usage="JUPYTER"),
        ModuleInstall(
            "parso", "pip", purpose="Parso is a Python parser that supports error recovery and " +
            "round-trip parsing for different Python versions (in multiple Python versions). Parso " +
            "is also able to list multiple syntax errors in your python file."),
        ModuleInstall(
            "jedi", "pip", purpose="An autocompletion tool for Python that can be used for text editors."),
        ModuleInstall("backcall", "pip", usage="JUPYTER",
                      purpose="Specifications for callback functions passed in to an API (IPython, Jupyter)"),
        ModuleInstall(
            "ipython", "pip", mname="IPython", purpose="IPython, Jupyter", usage="JUPYTER"),
        ModuleInstall(
            "jupyter_core", "pip", purpose="Jupyter Core", usage="JUPYTER"),
        ModuleInstall(
            "jupyter", "pip", purpose="Jupyter", usage="JUPYTER"),
        ModuleInstall(
            "jupyter_client", "pip", purpose="Jupyter client", usage="JUPYTER"),
        ModuleInstall(
            "jupyterlab_pygments", "pip", purpose="Jupyterlab", usage="JUPYTER"),
        ModuleInstall(
            "nbformat", "pip", purpose="IPython, notebooks conversion, new in Jupyter 4.0", usage="JUPYTER"),
        ModuleInstall(
            "entrypoints", "pip", purpose="Discover and load entry points from installed packages.", usage="JUPYTER"),
        ModuleInstall(
            "nbconvert", "pip", purpose="IPython, notebooks conversion, new in Jupyter 4.0", usage="JUPYTER"),
        ModuleInstall(
            "jupyter_sphinx", "pip", purpose="Jupyter Sphinx Extensions", usage="JUPYTER"),
        ModuleInstall(
            "jupyterlab-launcher", "pip", mname="jupyterlab_launcher",
            purpose="Jupyter Lab Launcher", usage="JUPYTER"),
        ModuleInstall("jupyterlab-server", "pip", mname="jupyterlab_server",
                      purpose="Jupyter Lab Server", usage="JUPYTER"),
        ModuleInstall(
            "jupyterlab", "pip", purpose="Jupyter Lab", usage="JUPYTER"),
        ModuleInstall(
            "ipympl", "pip", purpose="Matplotlib Jupyter Extension", usage="JUPYTER"),
        ModuleInstall(
            "nbsphinx", "pip", purpose="nbsphinx is a Sphinx extension that provides a source parser for *.ipynb files.", usage="JUPYTER"),
        ModuleInstall("PyPDF2", "pip", purpose="PDF toolkit", usage="PDF"),
        ModuleInstall(
            "ghost.py", "pip", mname="ghost", purpose="ghost.py is a webkit web client written in python", usage="JUPYTER"),
        ModuleInstall("nbbrowserpdf", "pip",
                      purpose="LaTeX-free PDF generation from Jupyter Notebooks", usage="JUPYTER"),
        ModuleInstall("Send2Trash", "pip", mname="send2trash",
                      purpose="Send file to trash natively under Mac OS X, Windows and Linux."),
        ModuleInstall(
            "notedown", "pip", purpose="Convert markdown to IPython notebook.", usage="JUPYTER"),
        ModuleInstall(
            "ipykernel", "pip", purpose="IPython, Jupyter, kernels", usage="JUPYTER"),
        ModuleInstall(
            "qtconsole", "pip", purpose="IPython, notebooks, qtconsole", usage="JUPYTER"),
        ModuleInstall("zipp", "pip",
                      purpose="A pathlib-compatible Zipfile object wrapper."),
        ModuleInstall("importlib-metadata", "pip", mname="importlib_metadata",
                      purpose="library to access the metadata for a Python package. It is intended to be ported to Python 3.8.",
                      usage="JUPYTER"),
        ModuleInstall(
            "path.py", "pip", mname="path", purpose="IPython, dependency", usage="JUPYTER"),
        ModuleInstall(
            "micropython-libc", "pip", mname="libc", purpose="dependency for ptyprocess, MicroPython FFI helper module",
            usage="JUPYTER/LINUX") if not sys.platform.startswith("win") else None,
        ModuleInstall(
            "micropython-ffilib", "pip", mname="ffi", purpose="dependency for ptyprocess, MicroPython FFI helper module",
            usage="JUPYTER/LINUX") if not sys.platform.startswith("win") else None,
        ModuleInstall(
            "micropython-fcntl", "pip", mname="fcntl", purpose="dependency for ptyprocess, Functions to compute fnctl.ioctl's opt argument",
            usage="JUPYTER/LINUX") if not sys.platform.startswith("win") else None,
        ModuleInstall(
            "ptyprocess", "pip", purpose="dependency for the terminado, Run a subprocess in a pseudo terminal",
            usage="JUPYTER/LINUX"),
        ModuleInstall("pexpect", "pip", usage="JUPYTER",
                      purpose="needed by ipykernel on Linux, Pexpect makes Python a better tool for controlling " +
                      "other applications (needed by metakernel)."),
        ModuleInstall("pywinpty", "wheel",
                      purpose="Python bindings for the winpty pseudo terminal library. It allows to create and " +
                      "communicate with Windows processes that print outputs and recieve inputs via console input " +
                      "and output pipes."),
        ModuleInstall(
            "terminado", "pip", purpose="dependency for the notebooks, Terminals served to term.js using Tornado websockets",
            usage="JUPYTER/LINUX"),
        ModuleInstall(
            "ipywidgets", "pip", purpose="IPython, Jupyter, widgets", usage="JUPYTER"),
        ModuleInstall("ipyscales", "pip",
                      purpose="A widget library for scales", usage="JUPYTER"),
        ModuleInstall("ipydatawidgets", "pip",
                      purpose="A set of widgets to help facilitate reuse of large datasets across widgets", usage="JUPYTER"),
        ModuleInstall(
            "backports_abc", "pip", purpose="A backport of recent additions to the 'collections.abc' module",
            usage="JUPYTER"),
        ModuleInstall("yapf", "pip", purpose="Code formatter"),
        ModuleInstall(
            "widgetsnbextension", "pip",
            purpose="Interactive HTML widgets for Jupyter notebooks.", usage="JUPYTER"),
        ModuleInstall(
            "jupyter_contrib_nbextensions", "github", "ipython-contrib",
            purpose="This repository contains a collection of extensions that add functionality to the Jupyter notebook.", usage="JUPYTER"),
        ModuleInstall(
            "backports.shutil-get-terminal-size", "pip", mname="backports", purpose="needed for Jupyter",
            usage="JUPYTER"),
        ModuleInstall(
            "prometheus_client", "pip", purpose="The official Python 2 and 3 client for Prometheus.", usage="JUPYTER"),
        ModuleInstall(
            "notebook", "pip", purpose="Jupyter notebooks, new in Jupyter 4.0", usage="JUPYTER"),
        ModuleInstall(
            "nbpresent", "pip", purpose="Next generation slides from Jupyter Notebooks", usage="JUPYTER"),
        ModuleInstall(
            "jupyter-console", "pip", mname="jupyter_console", purpose="Jupyter console, new in Jupyter 4.0", usage="JUPYTER"),
        ModuleInstall(
            "metakernel", "pip", purpose="more magic commands for Jupyter", usage="JUPYTER"),
        ModuleInstall('SQLAlchemy', 'wheel', mname='sqlalchemy',
                      purpose="model SQL queries as objects", usage="SQL"),
        ModuleInstall('sqlparse', 'pip', usage="SQL",
                      purpose="Non-validating SQL parser"),
        ModuleInstall(
            "ipystata", "pip", purpose="Jupyter kernel for Stata",
            usage="JUPYTER/PY2") if sys.version_info[0] == 2 else None,
        ModuleInstall("jupyter-pip", "pip", mname="jupyterpip",
                      purpose="Allows Jupyter notebook extension writers to make their extension pip installable!",
                      usage="JUPYTER"),
        ModuleInstall("portalocker", "pip", usage="JUPYTER",
                      purpose="Portalocker is a library to provide an easy API to file locking."),
        ModuleInstall("ipyparallel", "pip", usage="JUPYTER",
                      purpose="Interactive Parallel Computing with IPython"),
        # end of ipython
        #
        ModuleInstall("typing", "pip", purpose="Type Hints for Python") if sys.version_info[
            :2] < (3, 5) else None,
        ModuleInstall("typecheck-decorator", "pip", mname="typecheck",
                      purpose="verifies decorators at running time"),
        ModuleInstall("requests-cache", "pip", mname="requests_cache",
                      purpose="Persistent cache for requests library"),
        ModuleInstall("SIP", "pip", mname="sip",
                      usage="GUI", purpose="For PyQt5"),
        ModuleInstall("ete3", "pip", "http://etetoolkit.org/", usage="VIZ",
                      purpose="tree visualisation, ete3 does not work with PyQt4 (2017-11)."),
        ModuleInstall("PyQt5-sip", "pip", mname="PyQt5_sip", usage="GUI"),
        ModuleInstall("PyQt5", "pip", usage="GUI"),
        ModuleInstall("shiboken2", "pip", purpose="for PySide"),
        ModuleInstall("qtpy", "pip", usage="GUI",
                      purpose="single interface for QtPy4, 5, PySide"),
        ModuleInstall(
            "PySide2", "pip",
            purpose="The Qt for Python project aims to provide a complete port of the PySide module to Qt 5. ",
            usage="GUI"),
        ModuleInstall(
            "psutil", "wheel", purpose="cross-platform library for retrieving information " +
            "onrunning processes and system utilization (CPU, memory, disks, network)in Python."),  #
        ModuleInstall(
            "rope_py3k", "pip", mname="rope", purpose="refactoring library") if sys.version_info[0] >= 3 else None,  #
        ModuleInstall(
            "isort", "pip", purpose="A Python utility / library to sort Python imports."),
        ModuleInstall(
            "backports.functools_lru_cache", "pip",
            purpose="backports.functools_lru_cache"),
        ModuleInstall("lazy-object-proxy", "pip", mname="lazy_object_proxy",
                      purpose="A fast and thorough lazy object proxy"),
        ModuleInstall("wrapt", "wheel",
                      purpose="A Python module for decorators, wrappers and monkey patching."),
        ModuleInstall(
            "typed_ast", "pip",
            purpose="typed_ast is a Python 3 package that provides a Python 2.7 and Python 3 "
            "parser similar to the standard ast library."),
        ModuleInstall(
            "astroid", "pip",
            purpose="A abstract syntax tree for Python with inference support."),
        ModuleInstall(
            "pylint", "pip", purpose="statistics on Python script style"),  #
        ModuleInstall(
            "pythonqwt", "pip", purpose="Qt plotting widgets (Spyder)"),
        ModuleInstall("wurlitzer", "pip", purpose="for spyder"),
        ModuleInstall("spyder-kernels", "pip", mname="spyder_kernels",
                      purpose="Jupyter Kernels for the Spyder console"),
        ModuleInstall(
            "spyder", "pip", mname="spyderlib", purpose="scientific IDE"),
        ModuleInstall("dbfread", "pip", purpose="access DBase format"),
        ModuleInstall(
            "aenum", "pip", purpose="Advanced Enumerations (compatible with Python's stdlib Enum), NamedTuples, and NamedConstants"),
        ModuleInstall("dbf", "pip", purpose="access DBase format"),
        ModuleInstall(
            "xmltodict", "pip", purpose="Makes working with XML feel like you are working with JSON"),
        ModuleInstall(
            "ansiconv", "pip", purpose="A Python module for converting ANSI coded text and converts it to either plain text or HTML."),
        ModuleInstall(
            "ansi2html", "pip", purpose="Convert text with ANSI color codes to HTML"),
        ModuleInstall(
            "nodeenv", "pip", purpose="Node.js virtual environment builder"),
        ModuleInstall("greenlet", "wheel",
                      purpose="Greenlet allows lightweight in-process concurrent programming."),
        ModuleInstall(
            'werkzeug', 'pip', purpose="The Swiss Army knife of Python web development"),
        ModuleInstall('itsdangerous', 'pip',
                      purpose="Various helpers to pass trusted data to untrusted environments and back."),
        ModuleInstall('simplejson', 'wheel',
                      purpose="Simple, fast, extensible JSON encoder/decoder for Python"),
        ModuleInstall(
            'ijson', 'pip', purpose="Iterative JSON parser with a standard Python iterator interface"),
        ModuleInstall("click", "pip",
                      purpose="A simple wrapper around optparse for powerful command line utilities."),
        ModuleInstall('flask', 'pip', usage="NETWORK",
                      purpose="Flask is a microframework for Python based on Werkzeug, " +
                      "Jinja 2 and good intentions. And before you ask: It's BSD licensed!"),
        ModuleInstall('flask-sqlalchemy', 'pip',
                      mname='flask.ext.sqlalchemy', usage="NETWORK"),
        ModuleInstall('Flask-Login', 'pip', mname="flask_login",
                      usage="NETWORK"),
        ModuleInstall("Flask-Cors", "pip", mname="flask_cors",
                      purpose="A Flask extension for handling Cross Origin Resource Sharing (CORS), " +
                      "making cross-origin AJAX possible."),
        ModuleInstall('PyYAML', 'wheel', mname='yaml',
                      purpose=" YAML parser and emitter for Python"),
        ModuleInstall('python-mimeparse', 'pip',
                      purpose="A module provides basic functions for parsing mime-type names and matching " +
                      "them against a list of media-ranges. (falcon)"),
        ModuleInstall('falcon', 'pip', usage="NETWORK",
                      purpose="Falcon is a very fast, very minimal Python web framework for building microservices, " +
                      "app backends, and higher-level frameworks."),
        ModuleInstall('falcon-auth', 'pip', mname='falcon_auth', usage="NETWORK",
                      purpose="A falcon middleware + authentication backends that adds authentication layer to you app/api service."),
        ModuleInstall('waitress', 'pip', usage="NETWORK",
                      purpose="Waitress WSGI server"),
        ModuleInstall(
            'markdown', 'pip', purpose="markdown parser (for bokeh)"),
        ModuleInstall(
            'pystache', 'pip', purpose="Mustache for Python (for bokeh)"),
        ModuleInstall(
            'bokeh', 'pip', purpose="interactive graphs, zoomable, javascript", usage="VIZ"),
        ModuleInstall(
            'bkcharts', 'pip', purpose="High level chart types built on top of Bokeh", usage="VIZ"),
        ModuleInstall(
            'traittypes', 'pip', purpose="Custom trait types for scientific computing."),
        ModuleInstall(
            'bqplot', 'pip', purpose="interactive graphs, zoomable, d3.js for notebooks", usage="VIZ"),
        ModuleInstall(
            'seaborn', 'pip', purpose="nicer graphs than matplotlib for statistical purposes", usage="VIZ"),
        ModuleInstall(
            'scikit-plot', 'pip', purpose="nicer graphs than matplotlib for datascientist", usage="VIZ"),
        ModuleInstall(
            'toolz', 'pip', purpose="Toolz provides a set of utility functions for iterators, functions, and dictionaries.", usage="DATA/ML"),
        ModuleInstall(
            'cytoolz', 'wheel', purpose="Cython implementation of Toolz: High performance functional utilities", usage="DATA/ML"),
        ModuleInstall('snowballstemmer', 'pip',
                      purpose="This package provides 16 stemmer algorithms (15 + Porter English stemmer) generated from Snowball " +
                      "algorithms, needed by sphinx-rtd-theme."),
        ModuleInstall('sphinx-rtd-theme', 'pip', mname='sphinx_rtd_theme',
                      purpose="sphinx theme", usage="SPHINX"),
        ModuleInstall("feedparser", "pip", purpose="parse RSS streams"),
        ModuleInstall(
            "pbr", "pip", purpose="PBR is a library that injects some useful and sensible default behaviors into your setuptools run."),
        ModuleInstall(
            "multi-key-dict", "pip", mname="multi_key_dict", purpose="Multi key dictionary implementation"),
        ModuleInstall(
            "python-jenkins", "pip", mname="jenkins", purpose="interact with Jenkins"),
        ModuleInstall(
            'envoy', 'pip', purpose="Simple API for running external processes."),
        ModuleInstall('Logbook', 'wheel', mname='logbook',
                      purpose="A logging replacement for Python"),
        ModuleInstall(
            'pkginfo', 'pip', purpose="Query metadatdata from sdists / bdists / installed packages."),
        ModuleInstall("multipledispatch", "pip",
                      purpose="A relatively sane approach to multiple dispatch in Python."),
        ModuleInstall("future", "pip",
                      purpose="Clean single-source support for Python 3 and 2"),
        ModuleInstall("mock", "pip",
                      purpose="mock is a library for testing in Python. It allows you to replace parts of your system " +
                      "under test with mock objects and make assertions about how they have been used."),
        ModuleInstall("multimethods", "pip",
                      purpose="A multimethod implementation, loosely based on Guido’s initial ‘Five-minute Multimethods in Python."),
        ModuleInstall("distlib", "pip",
                      purpose="Low-level components of distutils2/packaging, augmented with higher-level APIs for "
                      "making packaging easier."),
        ModuleInstall("appdirs", "pip",
                      purpose="A small Python module for determining appropriate platform-specific dirs"),
        ModuleInstall("qgrid", "pip", usage="VIZ",
                      purpose="A Pandas DataFrame viewer for IPython Notebook."),
        ModuleInstall("ujson", "wheel",
                      purpose="Ultra fast JSON encoder and decoder for Python"),
        ModuleInstall("natsort", "pip", purpose="Sort lists naturally"),
        ModuleInstall("wget", "pip", purpose="pure python download utility"),
        ModuleInstall("queuelib", "pip",
                      purpose="Collection of persistent (disk-based) queues"),
        ModuleInstall("semantic_version", "pip",
                      purpose="A library implementing the 'SemVer' scheme."),
        ModuleInstall("unidecode", "pip",
                      purpose="ASCII transliterations of Unicode text"),
        ModuleInstall("sqlite_bro", "pip",
                      purpose="GUI for SQLite"),
        ModuleInstall("uvloop", "pip", purpose="Fast implementation of asyncio event loop on top of libuv") if sys.version_info[
            :2] > (3, 5) and not sys.platform.startswith("win") else None,  # requires VS 2015 ?
        ModuleInstall("tabulate", "pip",
                      purpose="Pretty-print tabular data"),
        ModuleInstall("httpie", "pip",
                      purpose="HTTPie - a CLI, cURL-like tool for humans"),
        ModuleInstall("version-information", "pip", mname="version_information",
                      purpose="Version information"),
        ModuleInstall("isodate", "pip",
                      purpose="An ISO 8601 date/time/duration parser and formatter"),
        ModuleInstall("dominate", "pip",
                      purpose="Dominate is a Python library for creating and manipulating HTML documents using an elegant DOM API."),
        ModuleInstall("mbstrdecoder", "pip",
                      purpose="multi-byte character string decoder"),
        ModuleInstall("ipaddress", "pip",
                      purpose="IPv4/IPv6 manipulation library"),
        ModuleInstall("typepy", "pip",
                      purpose="A python library for variable type checker/validator/converter at run time."),
        ModuleInstall("pathvalidate", "pip",
                      purpose="A python library to validate/sanitize a string such as filenames/variable-names/excel-sheet-names."),
        ModuleInstall("toml", "pip",
                      purpose="Python Library for Tom's Obvious, Minimal Language"),
        ModuleInstall("DataProperty", "pip", mname="dataproperty",
                      purpose="Python library for extract property from data."),
        ModuleInstall("tabledata", "pip",
                      purpose="A Python library to represent tabular data for pytablewriter/pytablereader/SimpleSQLite."),
        ModuleInstall("msgfy", "pip",
                      purpose="msgfy is a Python library for convert Exception instance to a human-readable error message."),
        ModuleInstall("SimpleSQLite", "pip", mname="simplesqlite",
                      purpose="SimpleSQLite is a Python library to simplify the table creation and data insertion into SQLite database."),
        ModuleInstall("sqliteschema", "pip",
                      purpose="sqliteschema is a Python library to dump table schema of a SQLite database file."),
        ModuleInstall("pytablereader", "pip",
                      purpose="A python library to load structured table data from files/URL with various data format: " +
                      "CSV/Excel/HTML/JSON/LTSV/Markdown/TSV."),
        ModuleInstall("elasticsearch", "pip",
                      purpose="Python client for Elasticsearch"),
        ModuleInstall("pytablewriter", "pip",
                      purpose="convert a dataframe into many formats"),
        ModuleInstall("defusedxml", "pip",
                      purpose="XML bomb protection for Python stdlib modules"),
        ModuleInstall(
            "tqdm", "pip", purpose="A Simple Python Progress Meter", usage="JUPYTER"),

        ModuleInstall(
            'olefile', 'pip', purpose="Python package to parse, read and write Microsoft OLE2 files " +
            "(Structured Storage or Compound Document, Microsoft Office) - Improved version of the " +
            "OleFileIO module from PIL, the Python Image Library."),
        ModuleInstall(
            'Pillow', 'wheel', mname='PIL', purpose="read/create images"),
        # Sphinx extension to draw graphs.
        ModuleInstall(
            "webcolors", "pip", purpose="A library for working with color names and color value formats defined " +
            "by the HTML and CSS specifications for use in documents on the Web."),
        ModuleInstall(
            "funcparserlib", "pip", purpose="Recursive descent parsing library based on functional combinators"),
        ModuleInstall(
            "blockdiag", "pip", purpose="blockdiag generates block-diagram image from text"),
        ModuleInstall(
            "sphinxcontrib-blockdiag", "pip", mname="sphinxcontrib.blockdiag",
            purpose="Sphinx 'blockdiag' extension"),
    ]

    return [_ for _ in mod if _ is not None]



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper import get_module_dependencies, get_module_metadata, version_consensus
from pymyinstall.installhelper.module_install_exceptions import WrongVersionError
from pymyinstall.installhelper.install_cmd_helper import is_conda_distribution
from pymyinstall.installhelper.pip_helper import get_installed_distributions


class TestModuleDependencies(unittest.TestCase):

    def common_function(self, name, use_pip=None):
        res = get_module_dependencies(name, deep=True, use_pip=use_pip)
        for k, v in sorted(res.items()):
            assert isinstance(v, tuple)
            fLOG(k, "-->", v)
        if len(res) < 3:
            pkgs = get_installed_distributions(
                local_only=False, skip=[])
            req_map = dict((p.key, (p, p.requires())) for p in pkgs)
            mat = req_map.get(name, None)
            raise AssertionError("len(res)={0}\nres={1}\ndata={2}\nmat={3}".format(
                len(res), res, get_module_metadata(name), mat))

    def test_dependencies_matplotlib(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        if not is_conda_distribution():
            self.common_function("matplotlib")
        else:
            # issue with Anaconda, not investigated
            return

    def test_version_consensus(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        self.assertEqual(version_consensus('>=2.3', '>=2.4'), '>=2.4')
        self.assertEqual(version_consensus('>=2.5', '>=2.4'), '>=2.5')
        self.assertEqual(version_consensus('>2.3', '>2.4'), '>2.4')
        self.assertEqual(version_consensus('>2.5', '>2.4'), '>2.5')
        self.assertEqual(version_consensus('>=2.3', '>2.4'), '>2.4')
        self.assertEqual(version_consensus('>2.3', '>=2.4'), '>=2.4')
        self.assertEqual(version_consensus('>2.5', '>2.4'), '>2.5')
        self.assertEqual(version_consensus('>=2.5', '>2.4'), '>=2.5')
        self.assertEqual(version_consensus('>2.5', '>=2.4'), '>2.5')

        self.assertEqual(version_consensus('<=2.3', '<=2.4'), '<=2.3')
        self.assertEqual(version_consensus('<=2.5', '<=2.4'), '<=2.4')
        self.assertEqual(version_consensus('<2.3', '<2.4'), '<2.3')
        self.assertEqual(version_consensus('<2.5', '<2.4'), '<2.4')
        self.assertEqual(version_consensus('<=2.3', '<2.4'), '<=2.3')
        self.assertEqual(version_consensus('<2.3', '<=2.4'), '<2.3')
        self.assertEqual(version_consensus('<2.5', '<2.4'), '<2.4')
        self.assertEqual(version_consensus('<=2.5', '<2.4'), '<2.4')
        self.assertEqual(version_consensus('<2.5', '<=2.4'), '<=2.4')

        self.assertEqual(version_consensus('==2.3', '<=2.4'), '==2.3')
        self.assertEqual(version_consensus('<=2.5', '==2.4'), '==2.4')
        self.assertEqual(version_consensus('==2.3', '<2.4'), '==2.3')
        self.assertEqual(version_consensus('<2.5', '==2.4'), '==2.4')
        self.assertEqual(version_consensus('>2.3', '==2.4'), '==2.4')
        self.assertEqual(version_consensus('==2.5', '>2.4'), '==2.5')
        self.assertEqual(version_consensus('==2.5', '>2.4'), '==2.5')
        self.assertEqual(version_consensus('==2.5', '>=2.4'), '==2.5')

        try:
            version_consensus('<=2.3', '>=2.4')
        except WrongVersionError:
            pass

        try:
            version_consensus('<=2.3', '==2.4')
        except WrongVersionError:
            pass


if __name__ == "__main__":
    # TestModuleDependencies().test_dependencies_matplotlib()
    unittest.main()



# -*- coding: utf-8 -*-
"""
@file
@brief Defines a set of module for teaching purpose.
"""
from ..installhelper.module_install import ModuleInstall


def teachings_set():
    """
    modules implemented for my teachings, it requires the modules in set *ml*
    """
    mod = [
        ModuleInstall(
            "jyquickhelper", "pip", purpose="Helpers for Jupyter notebooks.", usage="TEACH"),
        ModuleInstall(
            "pyquickhelper", "pip", purpose="helpers to generation documentation", usage="TEACH"),
        ModuleInstall(
            "tkinterquickhelper", "pip", purpose="windows on the top of tkinter", usage="TEACH"),
        ModuleInstall(
            "pymyinstall", "pip", purpose="easy installation of modules including Windows", usage="TEACH"),
        ModuleInstall("pymmails", "pip",
                      purpose="read/send emails", usage="TEACH"),
        ModuleInstall(
            "pyenbc", "pip", purpose="Helpers for remote work, clusters, hadoop", usage="TEACH"),
        ModuleInstall(
            "pyensae", "pip", purpose="helpers, Hadoop, SQL, financial times series, ...", usage="TEACH"),
        ModuleInstall("pyrsslocal", "pip",
                      purpose="RSS readers", usage="TEACH"),
        ModuleInstall(
            "code_beatrix", "pip", purpose="teaching programming to kids, lesenfantscodaient.fr", usage="TEACH"),
        ModuleInstall(
            "actuariat_python", "pip", purpose="teachings, insurance examples", usage="TEACH"),
        ModuleInstall("ensae_teaching_cs", "pip",
                      purpose="teachings, introduction to programming, machine learning, map/reduce", usage="TEACH"),
        ModuleInstall("jupytalk", "pip",
                      purpose="materials for presentations", usage="TEACH"),
        ModuleInstall("mlstatpy", "pip",
                      purpose="materials for machine learning", usage="TEACH"),
        ModuleInstall("teachpyx", "pip",
                      purpose="materials for teachings", usage="TEACH"),
        ModuleInstall("ensae_projects", "pip",
                      purpose="single use code", usage="TEACH"),
        ModuleInstall("lightmlrestapi", "pip",
                      purpose="lightmlrestapi implements a light machine learning REST API based on falcon.", usage="TEACH"),
        ModuleInstall("lightmlboard", "pip",
                      purpose="lightmlboard implements a light machine learning leaderboard based on tornado.", usage="TEACH"),
        ModuleInstall("pandas_streaming", "pip",
                      purpose="pandas_streaming aims at processing big files with pandas, too big to hold in memory," +
                      "too small to be parallelized with a significant gain. The module replicates a subset of pandas " +
                      "API and implements other functionalities for machine learning.", usage="TEACH"),
        ModuleInstall("manydataapi", "pip",
                      purpose="Helpers to access APIs", usage="TEACH"),
        ModuleInstall("cpyquickhelper", "wheel",
                      purpose="Experiments with C++ and Python", usage="TEACH"),
        ModuleInstall("mlinsights", "wheel",
                      purpose="mlinsights implements functions to get insights on machine learned models.", usage="TEACH"),
        ModuleInstall("mlprodict", "wheel",
                      purpose="mlprodict implements a couple of ways to productionize machine learning predictions.", usage="TEACH"),
        ModuleInstall("wrapclib", "wheel",
                      purpose="Wraps existing C libraries.", usage="TEACH"),
        ModuleInstall("csharpy", "wheel",
                      purpose="Python + C# + C", usage="TEACH"),
    ]
    #
    return [_ for _ in mod if _ is not None]



"""
@file
@brief Some fixes for deprecated functions and not updated packages.
"""


def fix_scipy10_for_statsmodels08():
    """
    :epkg:`scipy` deprecated some functions
    still used by :epkg:`statsmodels` 0.8.
    See `deprecated features <https://scipy.github.io/devdocs/release.1.0.0.html#deprecated-features>`_
    and pull request `3942 <https://github.com/statsmodels/statsmodels/pull/3942/files>`_
    in :epkg:`statsmodels`.
    """
    import scipy.stats
    if not hasattr(scipy.stats, "chisqprob"):
        import scipy.stats.distributions
        setattr(scipy.stats, "chisqprob", scipy.stats.distributions.chi2.sf)


if __name__ == "__main__":
    fix_scipy10_for_statsmodels08()



"""
@brief      test log(time=3s)
"""
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.packaged import small_set
from pymyinstall.win_installer.win_batch import create_win_batches


class TestBatch(unittest.TestCase):

    def test_batch(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        temp = get_temp_folder(__file__, "temp_batch")
        list_modules = small_set()
        folders = dict(tools=temp, config=temp)
        op = create_win_batches(
            folders, selection={"r", "julia"}, fLOG=fLOG, module_list=list_modules)
        for _ in op:
            fLOG(_)
        assert len(op) > 0
        for _, b in op:
            assert os.path.exists(b)


if __name__ == "__main__":
    unittest.main()



"""
@brief      test log(time=1s)
"""
import unittest
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import ExtTestCase
from pymyinstall import build_requirements
from pymyinstall.packaged import small_set


class TestRequirements (ExtTestCase):

    def test_update_module(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod_list = small_set()
        res = build_requirements(mod_list)
        lines = res.split("\n")
        fLOG(len(lines))
        for i, line in enumerate(lines):
            fLOG(i, line)
        self.assertIn('pep8', res)
        self.assertGreater(len(lines), 100)

    def test_sort(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        mod_list = small_set()
        mod_list.sort()
        build_requirements(mod_list)
        k = [_.usage for _ in mod_list if _.usage is not None]
        k2 = list(sorted(_.usage for _ in mod_list if _.usage is not None))
        fLOG(k)
        fLOG(k2)
        self.assertEqual(k, k2)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
"""
@brief      test log(time=45s)
"""
import textwrap
import unittest
from pymyinstall.installcustom.install_custom_python import _clean_err1


class TestClearErr(unittest.TestCase):

    def test_clean_err(self):
        err = textwrap.dedent("""
                /_testcapimodule.c: In function ‘PyInit__testcapi’:
                /_testcapimodule.c:6216:5: warning: ‘tp_print’ is deprecated [-Wdeprecated-declarations]
                     MyList_Type.tp_print = 0;
                     ^~~~~~~~~~~
                In file included from ./Include/object.h:746:0,
                                 from ./Include/pytime.h:6,
                                 from ./Include/Python.h:85,
                                 from /var/lib/jenkins/workspace/pymyinstall/tpath/temp_py38_minimal/Python-3.8.1/Modules/_testcapimodule.c:15:
                ./Include/cpython/object.h:260:30: note: declared here
                     Py_DEPRECATED(3.8) int (*tp_print)(PyObject *, FILE *, int);
                                              ^~~~~~~~
                ERROR: typepy 0.6.4 requires six<2.0.0,>=1.10.0, which is not installed.
                ERROR: tabledata 0.9.1 requires six<2.0.0,>=1.10.0, which is not installed.
                ERROR: mbstrdecoder 0.8.4 requires chardet<4,>=3.0.4, which is not installed.
                ERROR: google-auth 1.11.0 requires cachetools<5.0,>=2.0.0, which is not installed.
                ERROR: google-auth 1.11.0 requires pyasn1-modules>=0.2.1, which is not installed.
                ERROR: google-auth 1.11.0 requires rsa<4.1,>=3.1.4, which is not installed.
                ERROR: google-auth 1.11.0 requires six>=1.9.0, which is not installed.
                ERROR: dataproperty 0.43.3 requires six<2.0.0,>=1.10.0, which is not installed.
                ERROR: chainer 7.1.0 requires filelock, which is not installed.
                ERROR: chainer 7.1.0 requires numpy>=1.9.0, which is not installed.
                ERROR: chainer 7.1.0 requires protobuf>=3.0.0, which is not installed.
                ERROR: chainer 7.1.0 requires six>=1.9.0, which is not installed.
                ERROR: chainer 7.1.0 requires typing-extensions, which is not installed.
                """)
        err2 = _clean_err1(err)
        self.assertEqual(err2, None)


if __name__ == "__main__":
    unittest.main()
